# Default configuration for knowledge repositories

import re


# A function called to see whether a specified path is permitted in the repository
# Only enforced when creating/modifying posts. It should return the path as a standard
# unix path (virtual folders separated by '/' and the final node should end in '.kp').
# It should raise an exception if the provided path is is not permitted in the knowledge
# repository. A default implementation is provided using `path_patterns`, which
# can be provided more readily in a YAML configuration file.
def path_parse(repo, path):
    if not path.endswith('.kp'):
        path += '.kp'
    for pattern in repo.config.path_patterns:
        if re.match(pattern, path):
            return path
    raise ValueError(
        "Provided path '{path}' does not match any of the following patterns:\n" +
        '\n'.join("'{}': {}".format(pattern, desc) for pattern, desc in repo.config.path_patterns.items())
    )


# The accepted path patterns should be provided as a dictionary mapping regex
# patterns to descriptions of that pattern.
path_patterns = {
    '.*': "Any path is valid."
}


# A dictionary of aliases which point to knowledge posts. This allows you to alias posts
# which may be useful to forward old links to new locations, or to give certain posts
# precedence. It also allows the post path to violate the normal naming rules of posts.
# The only condition is that they cannot mask an existing post url, in which case the
# alias is silently ignored.
aliases = {}


# Postprocessors to apply when importing KnowledgePost objects into the repository.
# Note that KnowledgePost objects by default run 'extract_images' and 'format_checks'.
# Order is important. Should be a list of tuples, of form:
# ('name of postprocessor', {'init_kwarg': 'value'})
postprocessors = {}


# Usernames of users to keep informed of changes to the knowledge data repo
editors = []


# Function to check whether provided username is a valid username, and if not, mutate it
# such that it is. Should raise an exception if that is not possible, and otherwise
# return the parsed username. A default implementation is provided using
# `username_pattern`, which can be provided more readily in a YAML configuration
# file.
def username_parse(repo, username):
    if not re.match(repo.config.username_pattern, username):
        raise ValueError(
            "Username '{}' does not follow the required pattern: '{}'"
            .format(repo.config.username_pattern)
        )
    return username


# A regex pattern to which valid usernames must adhere
username_pattern = '.*'


# Function to convert a username to a person's proper name.  A default
# implementation is provided using `username_to_name_pattern`, which can be
# provided more readily in a YAML configuration file.
def username_to_name(repo, username):
    m = re.match(repo.config.username_to_name_pattern[0], username)
    return repo.config.username_to_name_pattern[1].format(**m.groupdict())


# A tuple/list of strings, the first being a regex with named groups, and the
# second being a format string with the group names available.
username_to_name_pattern = ('(?P<username>.*)', '{username}')


# Function to convert a username to a person's email. A default implementation
# is provided using `username_to_email_pattern`, which can be provided more
# readily in a YAML configuration file.
def username_to_email(repo, username):
    m = re.match(repo.config.username_to_email_pattern[0], username)
    return repo.config.username_to_email_pattern[1].format(**m.groupdict())


# A tuple/list of strings, the first being a regex with named groups, and the
# second being a format string with the group names available.
username_to_email_pattern = ('(?P<username>.*)', '{username}@example.com')


# Function to generate the web uri for a knowledge post at path `path`. If
# `None`, should return the web uri for the entire repository. Return `None` if
# a web uri does not exist. A default implementation is provided using
# `web_uri_base`, which can be provided more readily in a YAML configuration
# file.
def web_uri(repo, path=None):
    if repo.config.web_uri_base:
        if path:
            return '/'.join([repo.config.web_uri_base, 'post', path])
        return repo.config.web_uri_base
    return None


# The base url of the server hosting this repository.
web_uri_base = None


# If administrators of this knowledge repository want to suggest a specific
# knowledge_repo version requirement when interacting with the repository using
# the `knowledge_repo` script, they may do so here. Users can always work around
# this restriction by using the `--dev` flag to the `knowledge_repo` script. If
# the value supplied is a string starting with '!', it is taken to refer to a
# git tag or commit hash on the upstream Knowledge Repo repository, and the
# `knowledge_repo` script will clone the required  Knowledge Repo version and
# chain commands to it. Otherwise, it is interpreted as a pip-like version
# description (e.g. '==x.y.z', '>0.1.2<=0.8.0', etc), and the currently running
# version of the `knowledge_repo` library is checked at runtime.
required_tooling_version = None


# WARNING: ADVANCED!
# Function that is called on a Flask web app instance before it is launched
# This is useful if you need to add a route for your local deployment, or other
# equally invasive action. Not recommended unless you know exactly what you are doing.
# It must return a KnowledgeFlask app instance.
def prepare_app(repo, app):
    return app



from .common import KnowledgeDeployer


class FlaskDeployer(KnowledgeDeployer):

    _registry_keys = ['flask']

    def run(self, **kwargs):
        self.app.start_indexing()
        return self.app.run(
            debug=self.app.config['DEBUG'],
            host=self.host,
            port=self.port,
            threaded=self.app.check_thread_support(),
            **kwargs
        )



from __future__ import absolute_import
from __future__ import unicode_literals
import markdown
from markdown import Extension
from markdown.blockprocessors import BlockProcessor
from markdown.preprocessors import Preprocessor
from markdown.util import AtomicString, etree
from markdown.extensions import codehilite, toc
from markdown.inlinepatterns import Pattern

import re
import base64
import mimetypes

from ..converter import KnowledgePostConverter
from ..mapping import SubstitutionMapper

MARKDOWN_EXTENSIONS = ['extra',
                       'abbr',
                       'attr_list',
                       'def_list',
                       'fenced_code',
                       'footnotes',
                       'tables',
                       'admonition',
                       codehilite.CodeHiliteExtension(guess_lang=False),
                       'meta',
                       'sane_lists',
                       'smarty',
                       toc.TocExtension(baselevel=1),
                       'wikilinks',
                       'knowledge_repo.converters.html:KnowledgeMetaExtension',
                       'knowledge_repo.converters.html:MathJaxExtension',
                       'knowledge_repo.converters.html:IndentsAsCellOutput',
                       'knowledge_repo.converters.html:InlineSpanStyles']


class InlineSpanStyles(Extension):

    SPAN_PATTERN = r'\[([\s\S]*?)\]\{((?:\ ?.[^\,\}]+?)*?)\}'

    class SpanMatchHandler(Pattern):
        def handleMatch(self, m):
            # Extract information from markdown tag
            text = m.group(2)

            ids = [
                id
                for id in m.group(3).split(',')
                if id.startswith('#')
            ]
            id = ids[0] if ids else None

            class_names = [
                class_name[1:] if class_name.startswith('.') else class_name
                for class_name in m.group(3).split(' ')
                if class_name.startswith('.')
            ]

            # Generate HTML element for new span
            el = etree.Element('span')
            el.text = text
            if id:
                el.attrib['id'] = id
            if class_names:
                el.attrib['class'] = " ".join(class_names)
            return el

    def extendMarkdown(self, md, md_globals):
        span_matcher = self.SpanMatchHandler(self.SPAN_PATTERN)
        md.inlinePatterns['inline_span'] = span_matcher


class IndentsAsCellOutputPreprocessor(Preprocessor):
    """
    Ensure all indented blocks are followed by a blank line to allow html
    preprocessors to extract html elements (like scripts) properly.
    """

    def run(self, lines):
        in_block = False
        block_startable = True
        for i, line in enumerate(lines):
            if not line.startswith(' ' * self.markdown.tab_length):
                if in_block:
                    if line != "":
                        lines.insert(i, "")
                    in_block = False
                else:
                    block_startable = True if line == "" else False
            elif block_startable and not in_block:
                in_block = True
        return lines


class IndentsAsCellOutputProcessor(BlockProcessor):
    """ Process code blocks. """

    def test(self, parent, block):
        return block.startswith(' ' * self.tab_length)

    def run(self, parent, blocks):
        sibling = self.lastChild(parent)
        block = blocks.pop(0)

        block, theRest = self.detab(block)
        block = block.rstrip()

        block_is_html = False
        if "<div " in block or "</" in block or "<span " in block:
            block_is_html = True

        if (sibling is not None and sibling.tag == "div"):
            # The previous block was a code block. As blank lines do not start
            # new code blocks, append this block to the previous, adding back
            # linebreaks removed from the split into a list.

            block_is_html = block_is_html and not isinstance(sibling.text, AtomicString)

            block = u'\n'.join([sibling.text, block])
            output = sibling
        else:
            # This is a new codeblock. Create the elements and insert text.
            output = markdown.util.etree.SubElement(parent, 'div', {'class': 'code-output'})

        # If not HTML, add the `pre` class so that we know to render output as raw text
        if not block_is_html and 'pre' not in output.get('class', 'code-output'):
            output.set('class', ' '.join([output.get('class', ''), 'pre']))

        output.text = "{}\n".format(block) if block_is_html else AtomicString("{}\n".format(block))

        if theRest:
            # This block contained unindented line(s) after the first indented
            # line. Insert these lines as the first block of the master blocks
            # list for future processing.
            blocks.insert(0, theRest)


class IndentsAsCellOutput(Extension):

    def extendMarkdown(self, md, md_globals=None):
        md.preprocessors.add("code_isolation",
                             IndentsAsCellOutputPreprocessor(md),
                             "<html_block")
        md.parser.blockprocessors['code'] = IndentsAsCellOutputProcessor(md.parser)


class KnowledgeMetaPreprocessor(Preprocessor):
    """ Get Meta-Data. """

    def run(self, lines):
        """ Parse Meta-Data and store in Markdown.Meta. """
        cnt = 0
        for i, line in enumerate(lines):
            if line.strip() == '---':
                cnt = cnt + 1
            if cnt == 2:
                break
        return lines[i + 1:]


class KnowledgeMetaExtension(Extension):
    """ Meta-Data extension for Python-Markdown. """

    def extendMarkdown(self, md, md_globals=None):
        """ Add MetaPreprocessor to Markdown instance. """
        md.preprocessors.add("knowledge_meta",
                             KnowledgeMetaPreprocessor(md),
                             ">normalize_whitespace")


class MathJaxPattern(markdown.inlinepatterns.Pattern):

    def __init__(self):
        markdown.inlinepatterns.Pattern.__init__(self, r'(?<!\\)(\$\$?)(.+?)\2')

    def handleMatch(self, m):
        node = markdown.util.etree.Element('mathjax')
        node.text = markdown.util.AtomicString(m.group(2) + m.group(3) + m.group(2))
        return node


class MathJaxExtension(markdown.Extension):
    def extendMarkdown(self, md, md_globals=None):
        # Needs to come before escape matching because \ is pretty important in LaTeX
        md.inlinePatterns.add('mathjax', MathJaxPattern(), '<escape')


class HTMLConverter(KnowledgePostConverter):
    '''
    Use this as a template for new KnowledgePostConverters.
    '''
    _registry_keys = ['html']

    def init(self):
        self.kp_images = self.kp.read_images()

    def _render_markdown(self, skip_headers=False, images_base64_encode=True, urlmappers=[]):
        """
        Returns the `Markdown` instance as well as the rendered html output
        as a tuple: (`Markdown`, str).
        """
        # Copy urlmappers locally so we can modify it without affecting global
        # state
        urlmappers = urlmappers[:]
        if images_base64_encode:
            urlmappers.insert(0, self.base64_encode_image_mapper)

        # proxy posts are assumed to be embeddable links
        if 'proxy' in self.kp.headers:
            return None, '<a href="{0}">Linked Post</a>\n<iframe width=100% height=800 src="{0}"></iframe>'.format(self.kp.headers['proxy'].strip())

        html = ''
        if not skip_headers:
            html += self.render_headers()

        md = markdown.Markdown(extensions=MARKDOWN_EXTENSIONS)
        html += md.convert(self.kp.read())

        html = self.apply_url_remapping(html, urlmappers)

        return md, html

    def to_string(self, skip_headers=False, images_base64_encode=True, urlmappers=[]):
        """
        Renders the markdown as html
        """
        return self._render_markdown(
            skip_headers=skip_headers,
            images_base64_encode=images_base64_encode,
            urlmappers=urlmappers
        )[1]

    def apply_url_remapping(self, html, urlmappers):
        patterns = {
            'img': '<img.*?src=[\'"](?P<url>.*?)[\'"].*?>',
            'a': '<a.*?href=[\'"](?P<url>.*?)[\'"].*?>'
        }

        def urlmapper_proxy(name, match):
            for urlmapper in urlmappers:
                new_url = urlmapper(name, match.group('url'))
                if new_url is not None:
                    return match.group(0).replace(match.group('url'), new_url)
            return None

        return SubstitutionMapper(patterns=patterns, mappers=[urlmapper_proxy]).apply(html)

    # Utility methods
    def render_headers(self):
        headers = self.kp.headers

        headers['authors_string'] = ', '.join(headers.get('authors'))
        headers['tldr'] = markdown.Markdown(extensions=MARKDOWN_EXTENSIONS[
                                            :-1]).convert(headers['tldr'])
        headers['date_created'] = headers['created_at'].isoformat()
        headers['date_updated'] = headers['updated_at'].isoformat()

        header = """
<h1>{title}</h1>
<p id='metadata'>
<strong>Author</strong>: {authors_string} <br>
<strong>Date Created</strong>: {date_created}<br>
<strong>Date Updated</strong>: {date_updated}<br>
<strong>Tags</strong><text>: </text><br>
<strong>TLDR</strong>: {tldr}<br>
</p>
""".format(**headers)

        return header

    def base64_encode_image_mapper(self, tag, url):
        if tag == 'img':
            if url in self.kp_images:
                image_data = base64.b64encode(self.kp_images[url])
                image_mimetype = mimetypes.guess_type(url)[0]
                if image_mimetype is not None:
                    return 'data:{};base64, '.format(image_mimetype) + image_data.decode('utf-8')
        return None



from __future__ import print_function
from builtins import input

import os
import shutil
import logging
import re
import git
import socket
import time
from io import open

from knowledge_repo._version import __git_uri__
from ..post import KnowledgePost
from ..repository import KnowledgeRepository
from ..utils.exec_code import get_module_for_source
from ..utils.types import str_types
from ..utils.encoding import encode

logger = logging.getLogger(__name__)


class FolderKnowledgeRepository(KnowledgeRepository):
    _registry_keys = ['', 'file']

    TEMPLATES = {
        'README.md': os.path.abspath(os.path.join(os.path.dirname(__file__), '../templates', 'repository_readme.md')),
        '.knowledge_repo_config.yml': os.path.abspath(os.path.join(os.path.dirname(__file__), '../templates', 'repository_config.yml'))
    }

    @classmethod
    def create(cls, uri, embed_tooling=False):
        if uri.startswith('file://'):
            uri = uri[len('file://'):]
        path = os.path.abspath(uri)
        if not os.path.exists(path):
            os.makedirs(path)

        # Add README and configuration templates
        for filename, template in cls.TEMPLATES.items():
            target = os.path.join(path, filename)
            if not os.path.exists(target):
                shutil.copy(template, target)
            else:
                logger.warning("Not overriding existing file '{}'.".format(filename))
        return FolderKnowledgeRepository(path)

    @classmethod
    def from_uri(cls, uri, *args, **kwargs):
        """
        If this folder is actually a git repository, a `GitKnowledgeRepository`
        is returned instead, unless the folder knowledge repository is explicitly
        requested via the 'file://' protocol.
        """
        check_for_git = True
        if uri.startswith('file://'):
            check_for_git = False
            uri = uri[len('file://'):]
        if check_for_git and os.path.exists(os.path.join(uri, '.git')):
            from .gitrepository import GitKnowledgeRepository
            return GitKnowledgeRepository(uri, *args, **kwargs)
        return cls(uri, *args, **kwargs)

    def init(self, config='.knowledge_repo_config.yml', auto_create=False):
        self.auto_create = auto_create
        self.path = self.uri
        self.config.update(os.path.join(self.path, config))

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        assert isinstance(path, str), "The path specified must be a string."
        path = os.path.abspath(os.path.expanduser(path))
        if not os.path.exists(path):
            path = os.path.abspath(path)
            if self.auto_create:
                self.create(path)
            else:
                raise ValueError("Provided path '{}' does not exist.".format(path))
        self._path = path

    # ----------- Repository actions / state ------------------------------------
    @property
    def revision(self):
        return time.time()

    @property
    def status(self):
        return 'OK'

    @property
    def status_message(self):
        return 'OK'

    # ---------------- Post retrieval methods --------------------------------

    def _dir(self, prefix, statuses):
        posts = set()

        if self.PostStatus.PUBLISHED in statuses:

            for path, folders, files in os.walk(os.path.join(self.path, prefix or '')):

                # Do not visit hidden folders
                for folder in folders:
                    if folder.startswith('.'):
                        folders.remove(folder)

                posts.update(
                    os.path.join(os.path.relpath(path, start=self.path), folder)
                    for folder in folders if folder.endswith('.kp')
                )
                posts.update(
                    os.path.join(os.path.relpath(path, start=self.path), file)
                    for file in files if file.endswith('.kp')
                )

        for post in sorted([post[2:] if post.startswith('./') else post for post in posts]):
            yield post

    # ------------- Post submission / addition user flow ----------------------
    def _add_prepare(self, kp, path, update=False, **kwargs):
        pass

    def _add_cleanup(self, kp, path, update=False, **kwargs):
        pass

    def _submit(self, path=None, branch=None, force=False):
        pass  # Added posts are already submitted

    def _publish(self, path):  # Publish a post for general perusal
        pass  # Added posts are already published

    def _unpublish(self, path):  # unpublish a post for general perusal
        raise NotImplementedError

    def _accept(self, path):  # Approve to publish a post for general perusal
        pass

    def _remove(self, path, all=False):
        shutil.rmtree(os.path.join(self.path, path))

    # ------------ Knowledge Post Data Retrieval Methods -------------------------

    def _kp_uuid(self, path):
        try:
            return self._kp_read_ref(path, 'UUID')
        except:
            return None

    def _kp_path(self, path, rel=None):
        return KnowledgeRepository._kp_path(self, os.path.expanduser(path), rel=rel or self.path)

    def _kp_exists(self, path, revision=None):
        return os.path.exists(os.path.join(self.path, path))

    def _kp_status(self, path, revision=None, detailed=False, branch=None):
        return self.PostStatus.PUBLISHED

    def _kp_get_revision(self, path):
        # We use a 'REVISION' file in the knowledge post folder rather than using git
        # revisions because using git rev-parse is slow.
        try:
            return int(self._kp_read_ref(path, 'REVISION'))
        except:
            return 0

    def _kp_get_revisions(self, path):
        raise NotImplementedError

    def _kp_write_ref(self, path, reference, data, uuid=None, revision=None):
        path = os.path.join(self.path, path)
        if os.path.isfile(path):
            kp = KnowledgePost.from_file(path, format='kp')
            kp._write_ref(reference, data)
            kp.to_file(path, format='kp')
        else:
            ref_path = os.path.join(path, reference)
            ref_dir = os.path.dirname(ref_path)
            if not os.path.exists(ref_dir):
                os.makedirs(ref_dir)
            with open(ref_path, 'wb') as f:
                return f.write(data)

    def _kp_dir(self, path, parent=None, revision=None):  # TODO: Account for revision
        path = os.path.join(self.path, path)
        if os.path.isdir(path):
            if parent:
                path = os.path.join(path, parent)
            for dirpath, dirnames, filenames in os.walk(os.path.join(self.path, path)):
                for filename in filenames:
                    if dirpath == "" and filename == "REVISION":
                        continue
                    yield os.path.relpath(os.path.join(dirpath, filename), os.path.join(self.path, path))
        else:
            kp = KnowledgePost.from_file(path, format='kp')
            for reference in kp._dir(parent=parent):
                yield reference

    def _kp_has_ref(self, path, reference, revision=None):  # TODO: Account for revision
        path = os.path.join(self.path, path)
        if os.path.isdir(path):
            return os.path.isfile(os.path.join(path, reference))
        else:
            kp = KnowledgePost.from_file(path, format='kp')
            return kp._has_ref(reference)

    def _kp_diff(self, path, head, base):
        raise NotImplementedError

    def _kp_new_revision(self, path, uuid=None):
        self._kp_write_ref(path, "REVISION", encode(self._kp_get_revision(path) + 1))
        if uuid:
            self._kp_write_ref(path, "UUID", encode(uuid))

    def _kp_read_ref(self, path, reference, revision=None):
        path = os.path.join(self.path, path)
        if os.path.isdir(path):
            with open(os.path.join(self.path, path, reference), 'rb') as f:
                return f.read()
        else:
            kp = KnowledgePost.from_file(path, format='kp')
            return kp._read_ref(reference)



import logging

from .common import KnowledgeDeployer, get_app_builder

# The following subclasses of KnowledgeDeployer must be imported in order to be registered as a deployer and hence
# made accessible using `KnowledgeDeployer.using(..)`.
from .flask import FlaskDeployer
from .uwsgi import uWSGIDeployer

# Wrap the gunicorn deployer in a try/except block, as it has a hard dependency on gunicorn which does not work on
# non-POSIX systems, or if it is not installed.
try:
    from .gunicorn import GunicornDeployer
except ImportError:
    pass



import os
import sys
import datetime
import logging
import traceback
from builtins import str
from future.utils import raise_with_traceback
from flask import current_app, request
from flask_login import UserMixin
from flask_sqlalchemy import SQLAlchemy
from collections import defaultdict

from sqlalchemy import func, distinct, and_, select, UniqueConstraint

from knowledge_repo._version import __version__
from knowledge_repo.repository import KnowledgeRepository
from .proxies import current_user, current_repo, db_session
from .utils.models import unique_constructor
from .utils.search import get_keywords
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.ext.associationproxy import association_proxy

logger = logging.getLogger(__name__)

db = SQLAlchemy()


class IndexMetadata(db.Model):
    __tablename__ = 'index_metadata'

    id = db.Column(db.Integer, nullable=False, primary_key=True)
    type = db.Column(db.String(255), nullable=False)
    name = db.Column(db.String(512), nullable=False)
    value = db.Column(db.String(512), nullable=True)
    updated_at = db.Column(db.DateTime, default=datetime.datetime.utcnow, onupdate=datetime.datetime.utcnow)

    @classmethod
    def get(cls, type, name, default=None):
        m = db_session.query(IndexMetadata).filter(IndexMetadata.type == type).filter(IndexMetadata.name == name).first()
        if m is not None:
            return m.value
        return default

    @classmethod
    def set(cls, type, name, value):
        m = db_session.query(IndexMetadata).filter(IndexMetadata.type == type).filter(IndexMetadata.name == name).first()
        if m is not None:
            m.value = value
            m.updated_at = datetime.datetime.utcnow()
        else:
            m = IndexMetadata(type=type, name=name, value=value, updated_at=datetime.datetime.utcnow())
            db_session.add(m)

    @classmethod
    def get_last_update(cls, type, name):
        m = db_session.query(IndexMetadata).filter(IndexMetadata.type == type).filter(IndexMetadata.name == name).first()
        if m is not None:
            return m.updated_at
        return None


class PostAuthorAssoc(db.Model):
    __tablename__ = 'assoc_post_author'

    post_id = db.Column(db.Integer, db.ForeignKey("posts.id"), nullable=False, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), nullable=False, primary_key=True)
    order = db.Column(db.Integer)

    post = db.relationship('Post', lazy='joined')
    author = db.relationship('User', lazy='joined')


assoc_post_tag = db.Table(
    'assoc_post_tag',
    db.Model.metadata,
    db.Column('post_id', db.Integer, db.ForeignKey('posts.id')),
    db.Column('tag_id', db.Integer, db.ForeignKey('tags.id'))
)

assoc_post_group = db.Table(
    'assoc_post_group',
    db.Model.metadata,
    db.Column('post_id', db.Integer, db.ForeignKey('posts.id')),
    db.Column('group_id', db.Integer, db.ForeignKey('groups.id'))
)

assoc_group_user = db.Table(
    'assoc_group_user',
    db.Model.metadata,
    db.Column('group_id', db.Integer, db.ForeignKey('groups.id')),
    db.Column('user_id', db.Integer, db.ForeignKey('users.id'))
)


class Comment(db.Model):
    __tablename__ = 'comments'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    post_id = db.Column(db.Integer)
    text = db.Column(db.Text)
    type = db.Column(db.String(100), default='post')
    created_at = db.Column(db.DateTime, default=func.now())
    updated_at = db.Column(db.DateTime, default=func.now(), onupdate=func.now())


class ErrorLog(db.Model):
    __tablename__ = 'errorlog'

    id = db.Column(db.Integer, primary_key=True)
    function = db.Column(db.String(100))
    location = db.Column(db.String(255))
    message = db.Column(db.Text())
    traceback = db.Column(db.Text())
    version = db.Column(db.String(100), default=__version__)
    created_at = db.Column(db.DateTime, default=func.now())

    @classmethod
    def from_exception(cls, e):
        tb = sys.exc_info()[-1]
        filename, linenumber, function, code = traceback.extract_tb(sys.exc_info()[-1])[-1]
        filename = os.path.relpath(filename, os.path.join(os.path.dirname(__file__), '..'))
        return ErrorLog(
            function=function,
            location=u'{}:{}'.format(filename, linenumber),
            message=u'{}: {}'.format(e.__class__.__name__, u"; ".join(str(a) for a in e.args)),
            traceback=u"\n".join(traceback.format_tb(tb))
        )

    @classmethod
    def logged(cls, function):
        def wrapped(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except Exception as e:
                db_session.rollback()
                db_session.add(ErrorLog.from_exception(e))
                db_session.commit()
                raise_with_traceback(e)
        return wrapped


class PageView(db.Model):
    __tablename__ = 'pageviews'

    id = db.Column(db.Integer, primary_key=True)
    id_errorlog = db.Column(db.Integer)
    page = db.Column(db.String(512))
    endpoint = db.Column(db.String(255))
    user_id = db.Column(db.Integer)
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100))
    object_action = db.Column(db.String(100))
    ip_address = db.Column(db.String(64))
    created_at = db.Column(db.DateTime, default=func.now())
    version = db.Column(db.String(100), default=__version__)

    class logged(object):

        def __init__(self, route, object_extractor=None):
            self._route = route
            self._object_extractor = object_extractor

        def __getattr__(self, attr):
            return getattr(self._route, attr)

        def __call__(self, *args, **kwargs):
            if not current_app.config.get('INDEXING_ENABLED', True):
                return self._route(*args, **kwargs)

            log = PageView(
                page=request.full_path,
                endpoint=request.endpoint,
                user_id=current_user.id,
                ip_address=request.remote_addr,
                version=__version__
            )
            errorlog = None
            log.object_id, log.object_type, log.object_action, reextract_after_request = self.extract_objects(*args, **kwargs)
            db_session.add(log)  # Add log here to ensure pageviews are accurate

            try:
                return self._route(*args, **kwargs)
            except Exception as e:
                db_session.rollback()  # Ensure no lingering database changes remain after crashed route
                db_session.add(log)
                errorlog = ErrorLog.from_exception(e)
                db_session.add(errorlog)
                db_session.commit()
                raise_with_traceback(e)
            finally:
                # Extract object id and type after response generated (if requested) to ensure
                # most recent data is collected
                if reextract_after_request:
                    log.object_id, log.object_type, log.object_action, _ = self.extract_objects(*args, **kwargs)

                if errorlog is not None:
                    log.id_errorlog = errorlog.id
                db_session.add(log)
                db_session.commit()

        def object_extractor(self, extractor):
            self._object_extractor = extractor
            return self

        def extract_objects(self, *args, **kwargs):
            if self._object_extractor is None:
                return None, None, None, False
            try:
                object_info = self._object_extractor(*args, **kwargs)
            except Exception as e:
                logger.warning("Error using object extractor: " + str(e))
                object_info = {'id': (-1), 'type': None}
            assert isinstance(object_info, dict), "Object extractors must return a dictionary."
            assert len(set(['id', 'type']).difference(object_info.keys())) == 0 and len(set(object_info.keys()).difference(['id', 'type', 'action', 'may_change'])) == 0, "Object extractors must at least include the keys 'id' and 'type', and optionally 'action' and 'may_change'. Was provided with: {}".format(str(list(object_info.keys())))
            object_info = defaultdict(lambda: None, object_info)
            return object_info['id'], object_info['type'], object_info['action'], object_info['may_change'] or False


class Vote(db.Model):
    __tablename__ = 'votes'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100), default='post')
    created_at = db.Column(db.DateTime, default=func.now())
    updated_at = db.Column(db.DateTime, default=func.now(), onupdate=func.now())


@unique_constructor(
    lambda identifier: identifier,
    lambda query, identifier: query.filter(User.identifier == identifier)
)
class User(db.Model, UserMixin):
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    created_at = db.Column(db.DateTime, default=func.now())

    identifier = db.Column(db.String(500))  # Unique identifier across all login methods

    username = db.Column(db.String(500))  # Username used to log in (may differ from identifier)
    password = db.Column(db.String(500))  # Password for local logins

    name = db.Column(db.String(500))  # Name as determined by auth method
    preferred_name = db.Column(db.String(500))  # Name as determined by user preferences

    email = db.Column(db.String(500))  # Email address
    avatar_uri = db.Column(db.Text())  # Either external url or data uri
    active = db.Column(db.Boolean, default=True)

    last_login_at = db.Column(db.DateTime)  # Date of last login

    _posts_assoc = db.relationship("PostAuthorAssoc")
    posts = association_proxy('_posts_assoc', 'post')  # This property should not directly modified

    # Method overrides for the UserMixin class for flask_login
    @property
    def is_active(self):
        return self.active

    @property
    def is_authenticated(self):
        return True

    @property
    def is_anonymous(self):
        return False

    def get_id(self):
        return self.identifier

    can_logout = True

    # Other useful methods
    @property
    def format_name(self):
        return self.preferred_name or self.name or self.identifier

    @property
    def subscriptions(self):  # TODO: make attribute style naming
        """Get the subscriptions associated with a user.

        Return an array of strings of tag_names
        """
        subscriptions = (db.session.query(Subscription)
                         .filter(Subscription.user_id == self.id)
                         .all())
        out_subscriptions = []
        for s in subscriptions:
            if s.object_type == 'tag':
                tag_obj = (db.session.query(Tag)
                           .filter(Tag.id == s.object_id)
                           .first())
                if tag_obj:
                    full_name = tag_obj.name
                    out_subscriptions.append(full_name)
                else:
                    db.session.delete(s)
        return out_subscriptions

    @property
    def liked_posts(self):
        """
        :return: Posts that a user has liked
        :rtype: list
        """
        votes = (db.session.query(Vote)
                 .filter(Vote.user_id == self.id)
                 .all())
        post_ids = [vote.object_id for vote in votes]
        if len(post_ids) == 0:
            return []
        excluded_tags = current_app.config.get('EXCLUDED_TAGS', [])
        posts = (db.session.query(Post)
                 .filter(Post.id.in_(post_ids))
                 .filter(~Post.tags.any(Tag.name.in_(excluded_tags)))
                 .all())
        return posts


@unique_constructor(
    lambda name: name,
    lambda query, name: query.filter(Tag.name == name)
)
class Tag(db.Model):
    __tablename__ = 'tags'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(500))
    _description = db.Column('description', db.Text())
    created_at = db.Column(db.DateTime, default=func.now())

    @hybrid_property
    def description(self):
        if self._description:
            return self._description
        return u"All posts with tag '{}'.".format(self.name)

    @description.expression
    def description(self):
        raise NotImplementedError


class Subscription(db.Model):
    __tablename__ = 'subscriptions'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100))  # Currently just tag
    created_at = db.Column(db.DateTime, default=func.now())


class Post(db.Model):
    __tablename__ = 'posts'

    id = db.Column(db.Integer, primary_key=True)
    uuid = db.Column(db.String(100), unique=True)
    path = db.Column(db.String(512))
    project = db.Column(db.String(512), nullable=True)  # DEPRECATED
    repository = db.Column(db.String(512))
    revision = db.Column(db.Integer())

    title = db.Column(db.Text())
    subtitle = db.Column(db.Text())
    tldr = db.Column(db.Text)
    keywords = db.Column(db.Text)
    thumbnail = db.Column(db.Text())

    private = db.Column(db.Integer())

    created_at = db.Column(db.DateTime, default=func.now())
    updated_at = db.Column(db.DateTime, default=func.now())

    _authors_assoc = db.relationship("PostAuthorAssoc",
                                     order_by='PostAuthorAssoc.order',
                                     collection_class=ordering_list('order'),
                                     cascade="all, delete-orphan")
    _authors = association_proxy('_authors_assoc', 'author',
                                 creator=lambda author: PostAuthorAssoc(author=author),)

    @hybrid_property
    def authors(self):
        return self._authors

    @authors.setter
    def authors(self, authors):
        """
        Sets the tags of the post to the tags given in comma delimited string
        form in tags_string
        """
        user_objs = []

        for author in authors:
            if not isinstance(author, User):
                author = author.strip()
                author = User(identifier=author)
            user_objs.append(author)

        self._authors = user_objs

    @hybrid_property
    def authors_string(self):
        return u', '.join([author.format_name for author in self.authors])

    @authors_string.expression
    def authors_string(self):
        raise NotImplementedError

    _tags = db.relationship("Tag", secondary=assoc_post_tag, backref='posts',
                            lazy='subquery')

    @hybrid_property
    def tags(self):
        return self._tags

    @tags.setter
    def tags(self, tags):
        """
        Sets the tags of the post to the tags given in comma delimited string
        form in tags_string
        """
        tag_objs = []

        for tag in tags:
            if not isinstance(tag, Tag):
                tag = tag.strip()
                if tag[0] == "#":
                    tag = tag[1:]
                tag = Tag(name=tag)
            tag_objs.append(tag)

        self._tags = tag_objs

    @property
    def contains_excluded_tag(self):
        excluded_tags = current_app.config.get('EXCLUDED_TAGS', [])
        return any([tag.name in excluded_tags for tag in self.tags])

    _groups = db.relationship("Group", secondary=assoc_post_group, backref='posts',
                              lazy='subquery')

    @hybrid_property
    def groups(self):
        return self._groups

    @groups.setter
    def groups(self, groups):
        # given a list of group_names, we add it.
        group_objs = []

        for group in groups:
            if not isinstance(group, Group):
                group = Group(name=group.strip())
            group_objs.append(group)

        # create an implicit group, group_post.id, to add
        # single users to
        group = Group(name=":post_group_" + str(self.id))

        # this created group should have the author associated with it
        # so they can add people to the post
        group.users = self.authors
        group_objs.append(group)

        self._groups = group_objs

    _status = db.Column('status', db.Integer(), default=0)

    @hybrid_property
    def status(self):
        return current_repo.PostStatus(self._status or 0)

    @status.expression
    def status(self):
        return func.coalesce(self._status, 0)

    @status.setter
    def status(self, status):
        if status is None:
            self._status = None
        else:
            assert isinstance(status, KnowledgeRepository.PostStatus), "Status must be an instance of KnowledgeRepository.PostStatus.Status or None"
            self._status = status.value

    @hybrid_property
    def is_published(self):
        return self.status == current_repo.PostStatus.PUBLISHED

    @is_published.expression
    def is_published(self):
        return func.coalesce(self._status, 0) == current_repo.PostStatus.PUBLISHED.value

    _views = db.relationship("PageView", lazy='dynamic',
                             primaryjoin="and_(foreign(PageView.object_id)==Post.id, "
                                         "PageView.object_type=='post',"
                                         "PageView.object_action=='view')")

    @hybrid_property
    def views(self):
        return self._views.all()

    @hybrid_property
    def view_count(self):
        return self._views.count()

    @view_count.expression
    def view_count(self):
        return (select([func.count(PageView.id)])
                .where(PageView.object_id == self.id)
                .where(PageView.object_type == 'post')
                .label("view_count"))

    @hybrid_property
    def view_user_count(self):
        return (db.session.query(func.count(distinct(PageView.user_id)))
                          .filter(PageView.object_id == self.id)
                          .filter(PageView.object_type == 'post')
                          .scalar())

    @view_user_count.expression
    def view_user_count(self):
        return (select([func.count(distinct(PageView.user_id))])
                .where(PageView.object_id == self.id)
                .where(PageView.object_type == 'post')
                .label("view_user_count"))

    _votes = db.relationship("Vote", lazy='dynamic',
                             primaryjoin="and_(foreign(Vote.object_id)==Post.id, "
                                         "Vote.object_type=='post')")

    @hybrid_property
    def votes(self):
        return self._votes.all()

    @hybrid_property
    def vote_count(self):
        """ Given the path of a post, return the total likes """
        return self._votes.count()

    @vote_count.expression
    def vote_count(self):
        return (select([func.count(Vote.id)])
                .where(Vote.object_id == self.id)
                .where(Vote.object_type == 'post')
                .label("vote_count"))

    def vote_counted_for_user(self, user_id):
        return (db_session.query(Vote)
                          .filter(and_(Vote.object_id == self.id, Vote.object_type == 'post', Vote.user_id == user_id))
                          .first()) is not None

    _comments = db.relationship("Comment", lazy="dynamic",
                                primaryjoin="and_(foreign(Comment.post_id)==Post.id, "
                                            "Comment.type=='post')")

    @hybrid_property
    def comments(self):
        return self._comments.all()

    @hybrid_property
    def comment_count(self):
        """ Given the path of the a post, return the total comments """
        return self._comments.count()

    @comment_count.expression
    def comment_count(self):
        return (select([func.count(Comment.id)])
                .where(Comment.post_id == self.id)
                .where(Comment.object_type == 'post')
                .label("comments_count"))

    @property
    def kp(self):
        return current_repo.post(self.path)

    @property
    def text(self):
        return self.kp.read()

    def update_metadata_from_kp(self, kp):
        """
        :param kp: Maps fields of the model to values
        :type kp: KnowledgePost
        :param kp: Maps fields of the model to values
        :type kr: KnowledgeRepository
        :return: None
        :rtype: None
        """
        headers = kp.headers

        self.uuid = kp.uuid
        self.path = kp.path
        self.project = headers.get('project')
        self.repository = kp.repository_uri
        self.revision = kp.revision
        self.title = headers['title']
        self.subtitle = headers.get('subtitle')
        self.tldr = headers['tldr']
        self.authors = headers.get('authors', [])
        self.tags = headers.get('tags', [])
        self.keywords = get_keywords(self)
        self.thumbnail = kp.thumbnail_uri

        self.created_at = headers['created_at']
        self.updated_at = headers['updated_at']
        if self.created_at > self.updated_at:
            self.updated_at = self.created_at

        self.status = kp.status

        self.private = 0
        # we do this check so that no header (None) and False are treated the same
        if headers.get('private', ''):
            self.private = 1
            self.groups = headers.get('allowed_groups', [])


class Email(db.Model):

    __tablename__ = 'emails'

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer)
    trigger_id = db.Column(db.Integer)
    trigger_type = db.Column(db.String(100))
    object_id = db.Column(db.Integer)
    object_type = db.Column(db.String(100))
    sent_at = db.Column(db.DateTime, default=func.now())
    subject = db.Column(db.Text)
    text = db.Column(db.Text)


@unique_constructor(
    lambda name: name,
    lambda query, name: query.filter(Group.name == name)
)
class Group(db.Model):

    __tablename__ = 'groups'

    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(128), unique=True)

    _users = db.relationship("User", secondary=assoc_group_user, backref='users',
                             lazy='subquery')

    @hybrid_property
    def users(self):
        return self._users

    @users.setter
    def users(self, user_objs):
        self._users = self._users + user_objs



import os
import logging
import subprocess
import tempfile

from ..converter import KnowledgePostConverter


logger = logging.getLogger(__name__)


class RmdConverter(KnowledgePostConverter):
    _registry_keys = ['rmd']

    def from_file(self, filename, rebuild=True):
        Rmd_filename = filename
        if rebuild:
            tmp_fd, tmp_path = tempfile.mkstemp()
            os.close(tmp_fd)

            runcmd = (
                "Rscript --no-save --no-restore --slave -e \""
                "library(knitr);"
                "setwd('{wd}');"
                "knit('{fname}', '{target_path}', quiet=F)"
                "\""
                .format(
                    wd=os.path.abspath(os.path.dirname(filename)),
                    fname=os.path.abspath(filename),
                    target_path=tmp_path
                )
            )

            # Replace '\' with '\\' on Windows machines so R happy with filepath
            if os.name == 'nt':
                runcmd = runcmd.replace("\\", "\\\\")

            subprocess.check_output(runcmd, shell=True)
            Rmd_filename = tmp_path

        with open(Rmd_filename) as f:
            self.kp_write(f.read())
        self.kp.add_srcfile(filename)

        # Clean up temporary file
        if rebuild:
            os.remove(tmp_path)



"""Add ErrorLog table.

Revision ID: 9715822acf6c
Revises: 860cb49889a9
Create Date: 2017-02-27 10:15:42.372833

"""

# revision identifiers, used by Alembic.
revision = '9715822acf6c'
down_revision = '860cb49889a9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('errorlog',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('function', sa.String(length=100), nullable=True),
        sa.Column('location', sa.String(length=255), nullable=True),
        sa.Column('message', sa.Text(), nullable=True),
        sa.Column('traceback', sa.Text(), nullable=True),
        sa.Column('version', sa.String(length=100), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.add_column('pageviews', sa.Column('id_errorlog', sa.Integer(), nullable=True))
    op.drop_column('pageviews', 'error_message')


def downgrade():
    op.add_column('pageviews', sa.Column('error_message', sa.Text(), nullable=True))
    op.drop_column('pageviews', 'id_errorlog')
    op.drop_table('errorlog')



"""Add subtitle field to post database.

Revision ID: d15f6cac07e1
Revises: 009eafe4838f
Create Date: 2018-10-03 12:31:18.462880

"""

# revision identifiers, used by Alembic.
revision = 'd15f6cac07e1'
down_revision = '009eafe4838f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('posts', sa.Column('subtitle', sa.Text(), nullable=True))


def downgrade():
    op.drop_column('posts', 'subtitle')



from flask import current_app
from flask_login import current_user
from werkzeug.local import LocalProxy

__all__ = ['db_session', 'current_repo', 'current_user']

db_session = LocalProxy(lambda: current_app.db.session)
current_repo = LocalProxy(lambda: current_app.repository)



import flask
from flask import request, render_template, flash, redirect, url_for
from flask_login import login_user

from ..models import User
from ..utils.auth import is_safe_url

from ..auth_provider import KnowledgeAuthProvider


class DebugAuthProvider(KnowledgeAuthProvider):
    _registry_keys = ['debug']

    def prompt(self):
        if request.method == 'POST':
            user = request.form['username']
            login_user(User(identifier=user))

            flash('Logged in successfully.')

            next = request.args.get('next')
            # is_safe_url should check if the url is safe for redirects.
            # See http://flask.pocoo.org/snippets/62/ for an example.
            if not is_safe_url(next):
                return flask.abort(400)

            return redirect(next or url_for('index.render_feed'))
        return render_template('auth-login-form.html', skip_password=True)

    def get_user(self):
        return User(identifier=request.form['username'])



import os
import types
import inspect
import sys
import textwrap
from abc import abstractmethod
from future.utils import with_metaclass

import knowledge_repo
from knowledge_repo.utils.registry import SubclassRegisteringABCMeta


def get_app_builder(uris, debug, db_uri, config, **kwargs):
    def get_app():
        return (
            knowledge_repo.KnowledgeRepository
            .for_uri(uris)
            .get_app(db_uri=db_uri, debug=debug, config=config, **kwargs)
        )
    return get_app


class KnowledgeDeployer(with_metaclass(SubclassRegisteringABCMeta, object)):

    def __init__(self,
                 knowledge_builder,
                 host='0.0.0.0',
                 port=7000,
                 workers=4,
                 timeout=60):
        assert isinstance(knowledge_builder, (str, types.FunctionType)), \
            u"Unknown builder type {}".format(type(knowledge_builder))
        self.knowledge_builder = knowledge_builder
        self.host = host
        self.port = port
        self.workers = workers
        self.timeout = timeout

    @classmethod
    def using(cls, engine):
        if engine == 'gunicorn':
            if sys.platform == 'win32':
                raise RuntimeError(
                    "`gunicorn` deployer is not available for Windows. Please "
                    "use `uwsgi` or `flask` engines instead."
                )
            elif 'gunicorn' not in cls._registry:
                raise RuntimeError(
                    "`gunicorn` does not appear to be installed. Please "
                    "install it and try again."
                )
        return cls._get_subclass_for(engine)

    @property
    def builder_str(self):
        if isinstance(self.knowledge_builder, types.FunctionType):
            out = []
            for nl, cell in zip(self.knowledge_builder.__code__.co_freevars, self.knowledge_builder.__closure__):
                if isinstance(cell.cell_contents, str):
                    out.append(u'{} = "{}"'.format(nl, cell.cell_contents.replace('"', '\\"')))
                else:
                    out.append(u'{} = {}'.format(nl, cell.cell_contents))
            out.append(textwrap.dedent(inspect.getsource(self.knowledge_builder)))
            return u'\n'.join(out)
        return self.knowledge_builder

    @property
    def builder_func(self):
        if isinstance(self.knowledge_builder, str):
            knowledge_builder = 'def get_app():\n\t' + self.knowledge_builder.replace('\n', '\t') + '\n\treturn app'
            namespace = {}
            exec(knowledge_builder, namespace)
            return namespace['get_app']
        return self.knowledge_builder

    @property
    def app(self):
        try:
            self.__app
        except AttributeError:
            self.__app = self.builder_func()
        return self.__app

    def write_temp_files(self):
        import tempfile
        tmp_dir = tempfile.mkdtemp()
        tmp_path = os.path.join(tmp_dir, 'server.py')

        kr_path = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../..'))

        out = []
        out.append('import sys')
        out.append(u'sys.path.insert(0, "{}")'.format(kr_path))
        out.append('import knowledge_repo')

        out.append(self.builder_str)
        if not isinstance(self.knowledge_builder, str):
            out.append('app = %s()' % self.knowledge_builder.__name__)
        out.append('app.start_indexing()')

        with open(tmp_path, 'w') as f:
            f.write(u'\n'.join(out))

        return tmp_dir

    @abstractmethod
    def run(self):
        raise NotImplementedError



"""Add support for external authentication.

Revision ID: 009eafe4838f
Revises: 9715822acf6c
Create Date: 2017-07-23 02:16:39.493032

"""

# revision identifiers, used by Alembic.
revision = '009eafe4838f'
down_revision = '9715822acf6c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    # Migrate from `usernames` to `identifiers`
    op.alter_column('users', 'username', new_column_name='identifier', existing_type=sa.String(length=500))
    op.add_column('users', sa.Column('username', sa.String(length=500), nullable=True))

    # Add new fields
    op.add_column('users', sa.Column('password', sa.String(length=500), nullable=True))
    op.add_column('users', sa.Column('active', sa.Boolean(), nullable=True))
    op.add_column('users', sa.Column('name', sa.String(length=500), nullable=True))
    op.add_column('users', sa.Column('preferred_name', sa.String(length=500), nullable=True))
    op.add_column('users', sa.Column('avatar_uri', sa.Text(), nullable=True))
    op.add_column('users', sa.Column('email', sa.String(length=500), nullable=True))
    op.add_column('users', sa.Column('last_login_at', sa.DateTime(), nullable=True))


def downgrade():
    op.drop_column('users', 'username')
    op.alter_column('users', 'identifier', new_column_name='username', existing_type=sa.String(length=500))

    # Remove new fields
    op.drop_column('users', 'preferred_name')
    op.drop_column('users', 'password')
    op.drop_column('users', 'name')
    op.drop_column('users', 'last_login_at')
    op.drop_column('users', 'email')
    op.drop_column('users', 'active')
    op.drop_column('users', 'avatar_uri')



from __future__ import absolute_import

import logging
import multiprocessing
import os
import time
from builtins import str

from .proxies import db_session, current_repo, current_app
from .models import ErrorLog, Post, IndexMetadata
from .utils.emails import send_subscription_emails
from .utils.time import time_since

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Some global aliases to be used below for enhanced readability
LOCKED = CHECKED = '1'
UNLOCKED = '0'


def set_up_indexing_timers(app):
    if not app.config['INDEXING_ENABLED']:
        return False

    if app.check_thread_support(check_index=True, check_repositories=app.config['INDEXING_UPDATES_REPOSITORIES']):
        if os.environ['KNOWLEDGE_REPO_MASTER_UUID'] != app.uuid:
            logger.info("Not spawning index-sync timers for non-master application instance: {}".format(app.uuid))
            return

        logger.info("Spawning index-sync timers for master application instance: {}".format(app.uuid))

        def index_watchdog(app):
            while True:
                if not hasattr(app, 'sync_thread') or not app.sync_thread.is_alive():
                    logger.warning("Master indexing thread has died. Restarting...")
                    with app.app_context():
                        app.sync_thread = multiprocessing.Process(target=index_sync_loop, args=(app,))
                        app.sync_thread.start()
                time.sleep(app.config['INDEXING_TIMEOUT'])

        def index_sync_loop(app):
            current_app.db.engine.dispose()
            while True:
                with app.app_context():
                    update_index(check_timeouts=False)
                time.sleep(app.config['INDEXING_INTERVAL'])

        app.index_watchdog = multiprocessing.Process(target=index_watchdog, args=(app,))
        app.index_watchdog.start()
    else:
        @app.before_request
        def update_index_if_required():
            update_index()


def acquire_index_lock():
    if IndexMetadata.get('lock', 'master_check', None) is None:
        try:
            IndexMetadata.set('lock', 'master_check', current_app.uuid)
            db_session.commit()
            if (
                IndexMetadata.get('lock', 'index_master') == current_app.uuid or
                time_since(IndexMetadata.get_last_update('lock', 'index_master'), default=current_app.config["INDEXING_TIMEOUT"] + 1) > current_app.config["INDEXING_TIMEOUT"]
            ):
                IndexMetadata.set('lock', 'index_master', current_app.uuid)  # Update/set lock and update timestamp
                return True
        finally:
            IndexMetadata.set('lock', 'master_check', None)
            db_session.commit()
    return False


def is_indexing():
    timeout = current_app.config["INDEXING_TIMEOUT"]  # Default index timeout to 10 minutes (after which indexing will be permitted to run again)
    last_update = time_since_index()
    return IndexMetadata.get('lock', 'index', UNLOCKED) == LOCKED and last_update is not None and (last_update < timeout)


def time_since_index(human_readable=False):
    last_update = IndexMetadata.get_last_update('lock', 'index')
    ts = time_since(last_update, human_readable=human_readable)
    if human_readable:
        if is_indexing():
            return 'Currently indexing'
        if ts is None:
            return "Never"
    return ts


def time_since_index_check(human_readable=False):
    last_update = IndexMetadata.get_last_update('lock', 'index_master')
    if human_readable and last_update is None:
        return "Never"
    return time_since(last_update, human_readable=human_readable)


def get_indexed_revisions():
    indexed = {}
    for uri in current_repo.uris.values():
        indexed_revision = IndexMetadata.get('repository_revision', uri)
        indexed[uri] = indexed_revision
    return indexed


def index_due_for_update():
    if not current_app.config['INDEXING_ENABLED']:
        return False

    interval = current_app.config["INDEXING_INTERVAL"]
    seconds = time_since_index()
    seconds_check = time_since_index_check()

    if (seconds is not None and seconds_check is not None) and (seconds < interval or seconds_check < interval):
        return False

    return True


def index_up_to_date():
    for uri, revision in current_repo.revisions.items():
        indexed_revision = IndexMetadata.get('repository_revision', uri)
        if indexed_revision is None or indexed_revision < str(revision):
            return False
    return True


@ErrorLog.logged
def update_index(check_timeouts=True, force=False, reindex=False):

    if not current_app.config['INDEXING_ENABLED']:
        return False

    if check_timeouts and not index_due_for_update():
        return False

    is_index_master = acquire_index_lock()

    # Check for update to repositories if configured to do so
    if (
        current_app.config['INDEXING_UPDATES_REPOSITORIES'] and
        (is_index_master or current_app.config['INDEXING_UPDATES_REPOSITORIES_WITHOUT_LOCK'])
    ):
        current_repo.update()

    # Short-circuit if not the index master (unless force is True)
    if not is_index_master and not force or index_up_to_date():
        return False

    try:
        IndexMetadata.set('lock', 'index', LOCKED)
        db_session.commit()

        kr_dir = {kp.path: kp for kp in current_repo.posts()}
        kr_uuids = {kp.uuid: kp for kp in kr_dir.values()}
        posts = db_session.query(Post).all()

        for post in posts:

            # If UUID has changed, check if we can find it elsewhere in the repository, and if so update index path
            if post.uuid and ((post.path not in kr_dir) or (post.uuid != kr_dir[post.path].uuid)):
                if post.uuid in kr_uuids:
                    logger.info(u'Updating location of post: {} -> {}'.format(post.path, kr_uuids[post.uuid].path))
                    post.path = kr_uuids[post.uuid].path

            # If path of post no longer in directory, mark as unpublished
            if post.path not in kr_dir:
                logger.info(u'Recording unpublished status for post at {}'.format(post.path))
                post.status = current_repo.PostStatus.UNPUBLISHED
                continue

            # Update database according to current state of existing knowledge post and
            # remove from kp_dir. This means that when this loop finishes, kr_dir will
            # only contain posts which are new to the repo.
            kp = kr_dir.pop(post.path)

            # Update metadata of post if required
            if reindex or (kp.revision > post.revision or not post.is_published or kp.uuid != post.uuid):
                if kp.is_valid():
                    logger.info(u'Recording update to post at: {}'.format(kp.path))
                    post.update_metadata_from_kp(kp)
                else:
                    logger.warning(u'Update to post at "{}" is corrupt.'.format(kp.path))

        # Add the new posts that remain in kr_dir
        for kp_path, kp in kr_dir.items():
            if not kp.is_valid():
                logger.warning(u'New post at "{}" is corrupt.'.format(kp.path))
                continue
            logger.info(u'creating new post from path {}'.format(kp_path))
            post = Post()
            db_session.add(post)
            db_session.flush()  # (matthew) Fix groups logic so this is not necessary
            post.update_metadata_from_kp(kp)
            send_subscription_emails(post)

        # Record revision
        for uri, revision in current_repo.revisions.items():
            IndexMetadata.set('repository_revision', uri, str(revision))
    finally:
        IndexMetadata.set('lock', 'index', UNLOCKED)
        db_session.commit()



import sys
import platform
import types
import logging
import tabulate
import knowledge_repo

from flask import request, current_app, Blueprint, url_for, Response

from ..proxies import current_repo
from ..index import get_indexed_revisions, is_indexing


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


blueprint = Blueprint('debug', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/debug/versions', methods=['GET'])
def show_versions():
    version_keys = ['__version__', 'VERSION', 'version', '__VERSION__']

    def get_version(module):
        for key in version_keys:
            if hasattr(module, key):
                version = getattr(module, key)
                if isinstance(version, types.ModuleType):
                    version = get_version(version)
                return version

        return 'Unknown'

    versions = []
    for module_name in sorted(set(name.split('.')[0] for name in sys.modules.keys())):
        if module_name.startswith('_'):
            continue
        module = __import__(module_name)
        if not hasattr(module, '__file__') or 'site-packages' not in module.__file__:
            continue
        versions.append([module_name, get_version(__import__(module_name))])

    # Show revision for mountpoint. Don't show actual uri because database passwords will be leaked
    revisions = []
    repo_revisions = current_repo.revisions
    indexed_revisions = get_indexed_revisions()
    for mountpoint, uri in current_repo.uris.items():
        revisions.append([mountpoint or '&lt;root&gt;', repo_revisions[uri], indexed_revisions.get(uri, 'None')])

    return ("Knowledge Repo Version: {}<br/>\n".format(knowledge_repo.__version__) +
            "Python Version: {}<br/>\n".format(sys.version) +
            "Platform: {}<br/>\n".format(platform.version()) +
            "<h2>Repository Revisions</h2>" +
            ("<i>Currently indexing</i><br />\n" if is_indexing() else "") +
            tabulate.tabulate(revisions, tablefmt='html') +
            "<h2>Loaded Packages</h2>\n" +
            tabulate.tabulate(versions, tablefmt='html', headers=('Mount Point', 'Revision', 'Indexed Revision')))


@blueprint.route('/debug/force_reindex', methods=['GET'])
def force_reindex():
    reindex = bool(request.args.get('reindex', ''))
    current_app.db_update_index(check_timeouts=False, force=True, reindex=reindex)
    return "Index Updated"


@blueprint.route('/debug/views')
def show_views():
    import urllib
    output = []
    for rule in current_app.url_map.iter_rules():
        options = {}
        for arg in rule.arguments:
            options[arg] = "[{0}]".format(arg)

        methods = ','.join(rule.methods)
        url = url_for(rule.endpoint, **options)
        line = urllib.unquote("{:50s} {:20s} {}".format(rule.endpoint, methods, url))
        output.append(line)

    return "<br />".join(sorted(output))


@blueprint.route('/debug/headers')
def show_headers():
    return Response(str(request.headers).encode(), mimetype='text/plain')



import datetime


def time_since(dt, human_readable=False, default=None):
    if dt is None:
        return default
    assert isinstance(dt, datetime.datetime)
    delta = (datetime.datetime.utcnow() - dt).total_seconds()
    if human_readable:
        return human_readable_time_delta(delta)
    return delta


def human_readable_time_delta(seconds):
    if seconds is None:
        return "Never"
    elif seconds < 60:
        return "{:d} seconds ago".format(int(round(seconds)))
    elif seconds < 60 * 60:
        return "{:d} minutes ago".format(int(round(seconds / 60)))
    elif seconds < 24 * 60 * 60:
        return "{:d} hours ago".format(int(round(seconds / 60 / 60)))
    else:
        return "{:d} days ago".format(int(round(seconds / 60 / 60 / 24)))



"""Functions that deal with getting the "keywords" for typeahead search

  Ultimately, this file will include all functions related to search. At the moment,
  it has two purposes:
    1. To hardcode the list of nltk stopwords in English
    2. Create a list of keywords from a list of words
"""
ENGLISH_STOPWORDS = [u'i', u'me', u'my', u'myself', u'we', u'our', u'ours', u'ourselves',
                     u'you', u'your', u'yours', u'yourself', u'yourselves', u'he', u'him',
                     u'his', u'himself', u'she', u'her', u'hers', u'herself', u'it', u'its',
                     u'itself', u'they', u'them', u'their', u'theirs', u'themselves', u'what',
                     u'which', u'who', u'whom', u'this', u'that', u'these', u'those', u'am',
                     u'is', u'are', u'was', u'were', u'be', u'been', u'being', u'have', u'has',
                     u'had', u'having', u'do', u'does', u'did', u'doing', u'a', u'an', u'the',
                     u'and', u'but', u'if', u'or', u'because', u'as', u'until', u'while',
                     u'of', u'at', u'by', u'for', u'with', u'about', u'against', u'between',
                     u'into', u'through', u'during', u'before', u'after', u'above', u'below',
                     u'to', u'from', u'up', u'down', u'in', u'out', u'on', u'off', u'over',
                     u'under', u'again', u'further', u'then', u'once', u'here', u'there',
                     u'when', u'where', u'why', u'how', u'all', u'any', u'both', u'each',
                     u'few', u'more', u'most', u'other', u'some', u'such', u'no', u'nor',
                     u'not', u'only', u'own', u'same', u'so', u'than', u'too', u'very',
                     u's', u't', u'can', u'will', u'just', u'don', u'should', u'now']


def get_keywords(post):
    " Given a post object return a space-separated string of keywords "
    author = [author.format_name for author in post.authors]
    title = post.title.split(" ")
    tldr = post.tldr.split(" ")
    tags = [tag.name for tag in post.tags]
    full_search_text = author + title + tldr + tags
    keywords = u' '.join([word for word in full_search_text if word not in ENGLISH_STOPWORDS])
    return keywords



"""Add index metadata table for on-demand reindexing.

Revision ID: 860cb49889a9
Revises: 36baffc5df12
Create Date: 2016-11-08 00:56:21.810049

"""

# revision identifiers, used by Alembic.
revision = '860cb49889a9'
down_revision = '36baffc5df12'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('index_metadata',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('type', sa.String(length=255), nullable=False),
    sa.Column('name', sa.String(length=512), nullable=False),
    sa.Column('value', sa.String(length=512), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('type', 'name', name='_uc_type_name')
    )


def downgrade():
    op.drop_table('index_metadata')



import sys

import markdown
from markdown.extensions import toc

import pygments
from flask import url_for
from jinja2 import Template
from knowledge_repo.post import KnowledgePost

MARKDOWN_EXTENSIONS = ['extra',
                       'abbr',
                       'attr_list',
                       'def_list',
                       'fenced_code',
                       'footnotes',
                       'tables',
                       'admonition',
                       'codehilite',
                       'meta',
                       'sane_lists',
                       'smarty',
                       toc.TocExtension(baselevel=1),
                       'wikilinks',
                       'nl2br']


def render_post_tldr(post):
    if isinstance(post, KnowledgePost):
        return markdown.Markdown(extensions=MARKDOWN_EXTENSIONS).convert(post.headers.get('tldr').strip())
    else:
        return markdown.Markdown(extensions=MARKDOWN_EXTENSIONS).convert(post.tldr.strip())


def render_post_header(post):

    header_template = Template(u"""
    <div class='metadata'>
    <span class='title'>{{title}}</span>
    {% if subtitle %}<span class='subtitle'>{{subtitle}}</span>{% endif %}
    <span class='authors'>{{authors}}</span>
    <span class='date_created'>{{date_created}}</span>
    <span class='date_updated'>(Last Updated: {{date_updated}})</span>
    <span class='tldr'>{{tldr}}</span>
    <span class='tags'></span>
    </div>
    """)

    def get_authors(usernames, authors):
        authors = [u"<a href='{}'>{}</a>".format(url_for('index.render_feed', authors=username), author) for username, author in zip(usernames, authors)]
        return u' and '.join(u', '.join(authors).rsplit(', ', 1))

    if isinstance(post, KnowledgePost):
        return header_template.render(title=post.headers['title'],
                                      subtitle=post.headers.get('subtitle'),
                                      authors=get_authors(post.headers['authors'], post.headers['authors']),
                                      date_created=post.headers['created_at'].strftime("%B %d, %Y"),
                                      date_updated=post.headers['updated_at'].strftime("%B %d, %Y"),
                                      tldr=render_post_tldr(post))
    else:
        return header_template.render(title=post.title,
                                      subtitle=post.subtitle,
                                      authors=get_authors([author.identifier for author in post.authors], [author.format_name for author in post.authors]),
                                      date_created=post.created_at.strftime("%B %d, %Y"),
                                      date_updated=post.updated_at.strftime("%B %d, %Y"),
                                      tldr=render_post_tldr(post))


def render_post_raw(post):
    if isinstance(post, KnowledgePost):
        raw_post = post.read().encode('ascii', 'ignore')
    else:
        raw_post = post.text.encode('ascii', 'ignore')

    raw_post = pygments.highlight(
        code=raw_post,
        lexer=pygments.lexers.get_lexer_by_name('md'),
        formatter=pygments.formatters.get_formatter_by_name('html')
    )

    return raw_post


def render_post(post, with_toc=False):
    """
    Renders the markdown as html
    """
    from knowledge_repo.converters.html import HTMLConverter

    def intra_knowledge_urlmapper(name, url):
        if name == 'a' and url.startswith('knowledge:'):
            return url_for('posts.render', path=url.split('knowledge:')[1]).replace('%2F', '/')  # Temporary fix before url revamp
        return None

    md, html = HTMLConverter(post if isinstance(post, KnowledgePost) else post.kp)._render_markdown(skip_headers=True, urlmappers=[intra_knowledge_urlmapper])

    html = render_post_header(post) + html

    if with_toc:
        return {
            "html": html,
            "toc": md.toc if md is not None else None
        }
    return html


def render_comment(comment):
    return markdown.Markdown().convert(comment.text)



""" Define the routes that show all the posts.

This includes:
  - /feed
  - /cluster
  - /table
  - /favorites
"""
import os
import json
from builtins import str
from collections import namedtuple
from flask import request, render_template, redirect, Blueprint, current_app, make_response
from flask_login import login_required
from sqlalchemy import case, desc

from .. import permissions
from ..proxies import db_session, current_repo
from ..utils.posts import get_posts
from ..models import Post, Tag, User, PageView
from ..utils.requests import from_request_get_feed_params
from ..utils.render import render_post_tldr

blueprint = Blueprint(
    'index', __name__, template_folder='../templates', static_folder='../static')


def has_no_empty_params(rule):
    defaults = rule.defaults if rule.defaults is not None else ()
    arguments = rule.arguments if rule.arguments is not None else ()
    return len(defaults) >= len(arguments)


@blueprint.route("/site-map")
@PageView.logged
def site_map():
    links = []
    for rule in current_app.url_map.iter_rules():
        # Filter out rules we can't navigate to in a browser
        # and rules that require parameters
        # if "GET" in rule.methods and has_no_empty_params(rule):
        # url = url_for(rule.endpoint, **(rule.defaults or {}))
        links.append((str(rule), rule.endpoint))
    # links is now a list of url, endpoint tuples
    return u'<br />'.join(str(link) for link in links)


@blueprint.route('/')
@PageView.logged
def render_index():
    return redirect('/feed')


@blueprint.route('/favorites')
@PageView.logged
@login_required
def render_favorites():
    """ Renders the index-feed view for posts that are liked """

    feed_params = from_request_get_feed_params(request)
    user_id = feed_params['user_id']

    user = (db_session.query(User)
            .filter(User.id == user_id)
            .first())
    posts = user.liked_posts

    post_stats = {post.path: {'all_views': post.view_count,
                              'distinct_views': post.view_user_count,
                              'total_likes': post.vote_count,
                              'total_comments': post.comment_count} for post in posts}

    return render_template("index-feed.html",
                           feed_params=feed_params,
                           posts=posts,
                           post_stats=post_stats,
                           top_header='Favorites')


@blueprint.route('/feed')
@PageView.logged
@permissions.index_view.require()
def render_feed():
    """ Renders the index-feed view """
    feed_params = from_request_get_feed_params(request)
    posts, post_stats = get_posts(feed_params)
    for post in posts:
        post.tldr = render_post_tldr(post)

    return render_template("index-feed.html",
                           feed_params=feed_params,
                           posts=posts,
                           post_stats=post_stats,
                           top_header='Knowledge Feed')


@blueprint.route('/table')
@PageView.logged
@permissions.index_view.require()
def render_table():
    """Renders the index-table view"""
    feed_params = from_request_get_feed_params(request)
    posts, post_stats = get_posts(feed_params)
    # TODO reference stats inside the template
    return render_template("index-table.html",
                           posts=posts,
                           post_stats=post_stats,
                           top_header="Knowledge Table",
                           feed_params=feed_params)


@blueprint.route('/cluster')
@PageView.logged
@permissions.index_view.require()
def render_cluster():
    """ Render the cluster view """
    # we don't use the from_request_get_feed_params because some of the
    # defaults are different
    filters = request.args.get('filters', '')
    sort_by = request.args.get('sort_by', 'alpha')
    group_by = request.args.get('group_by', 'folder')
    request_tag = request.args.get('tag')
    sort_desc = not bool(request.args.get('sort_asc', ''))

    excluded_tags = current_app.config.get('EXCLUDED_TAGS', [])
    post_query = (db_session.query(Post)
                            .filter(Post.is_published)
                            .filter(~Post.tags.any(Tag.name.in_(excluded_tags))))

    if filters:
        filter_set = filters.split(" ")
        for elem in filter_set:
            elem_regexp = "%," + elem + ",%"
            post_query = post_query.filter(Post.keywords.like(elem_regexp))

    ClusterPost = namedtuple(
        'ClusterPost',
        ['name', 'is_post', 'children_count', 'content']
    )

    if group_by == "author":
        author_to_posts = {}
        authors = (db_session.query(User).all())
        for author in authors:
            author_posts = [
                ClusterPost(name=post.title, is_post=True,
                            children_count=0, content=post)
                for post in author.posts
                if post.is_published and not post.contains_excluded_tag
            ]
            if author_posts:
                author_to_posts[author.format_name] = author_posts
        grouped_data = [
            ClusterPost(name=k, is_post=False,
                        children_count=len(v), content=v)
            for (k, v) in author_to_posts.items()
        ]

    elif group_by == "tags":
        tags_to_posts = {}
        all_tags = (db_session.query(Tag)
                              .filter(~Tag.name.in_(excluded_tags))
                              .all())

        for tag in all_tags:
            tag_posts = [
                ClusterPost(name=post.title, is_post=True,
                            children_count=0, content=post)
                for post in tag.posts
                if post.is_published and not post.contains_excluded_tag
            ]
            if tag_posts:
                tags_to_posts[tag.name] = tag_posts
        grouped_data = [
            ClusterPost(name=k, is_post=False,
                        children_count=len(v), content=v)
            for (k, v) in tags_to_posts.items()
        ]

    elif group_by == "folder":
        posts = post_query.all()

        # group by folder
        folder_to_posts = {}

        for post in posts:
            folder_hierarchy = post.path.split('/')
            cursor = folder_to_posts

            for folder in folder_hierarchy[:-1]:
                if folder not in cursor:
                    cursor[folder] = {}
                cursor = cursor[folder]

            cursor[folder_hierarchy[-1]] = post

        def unpack(d):
            """
            Recusively unpack folder_to_posts
            """
            children = []
            count = 0
            for k, v in d.items():
                if isinstance(v, dict):
                    l, contents = unpack(v)
                    count += l
                    children.append(
                        ClusterPost(name=k, is_post=False,
                                    children_count=l, content=contents)
                    )
                else:
                    count += 1
                    children.append(
                        ClusterPost(name=k, is_post=True,
                                    children_count=0, content=v)
                    )
            return count, children

        _, grouped_data = unpack(folder_to_posts)

    else:
        raise ValueError(u"Group by `{}` not understood.".format(group_by))

    def rec_sort(content, sort_by):
        sorted_content = []
        for c in content:
            if c.is_post:
                sorted_content.append(c)
            else:
                sorted_content.append(ClusterPost(
                    name=c.name,
                    is_post=c.is_post,
                    children_count=c.children_count,
                    content=rec_sort(c.content, sort_by)
                ))
        # put folders above posts
        clusters = [c for c in sorted_content if not c.is_post]
        posts = [c for c in sorted_content if c.is_post]
        if sort_by == "alpha":
            return (
                sorted(clusters, key=lambda x: x.name) +
                sorted(posts, key=lambda x: x.name)
            )
        else:
            return (
                sorted(clusters, key=lambda x: x.children_count, reverse=sort_desc) +
                sorted(posts, key=lambda x: x.children_count, reverse=sort_desc)
            )

    grouped_data = rec_sort(grouped_data, sort_by)

    return render_template("index-cluster.html",
                           grouped_data=grouped_data,
                           filters=filters,
                           sort_by=sort_by,
                           group_by=group_by,
                           tag=request_tag)


@blueprint.route('/create')
@blueprint.route('/create/<knowledge_format>')
@PageView.logged
@permissions.post_view.require()
def create(knowledge_format=None):
    """ Renders the create knowledge view """
    if knowledge_format is None:
        return render_template("create-knowledge.html",
                               web_editor_enabled=current_app.config['WEB_EDITOR_PREFIXES'] != [])

    cur_dir = os.path.dirname(os.path.realpath(__file__))
    knowledge_template = "knowledge_template.{}".format(knowledge_format)
    filename = os.path.join(cur_dir, '../../templates', knowledge_template)
    response = make_response(open(filename).read())
    response.headers["Content-Disposition"] = "attachment; filename=" + knowledge_template
    return response


@blueprint.route('/ajax/index/typeahead', methods=['GET', 'POST'])
def ajax_post_typeahead():
    if not permissions.index_view.can():
        return '[]'

    # this a string of the search term
    search_terms = request.args.get('search', '')
    search_terms = search_terms.split(" ")
    case_statements = []
    for term in search_terms:
        case_stmt = case([(Post.keywords.ilike('%' + term.strip() + '%'), 1)], else_=0)
        case_statements += [case_stmt]

    match_score = sum(case_statements).label("match_score")

    posts = (db_session.query(Post, match_score)
                       .filter(Post.status == current_repo.PostStatus.PUBLISHED.value)
                       .order_by(desc(match_score))
                       .limit(5)
                       .all())

    matches = []
    for (post, count) in posts:
        authors_str = [author.format_name for author in post.authors]
        typeahead_entry = {'author': authors_str,
                           'title': str(post.title),
                           'path': str(post.path),
                           'keywords': str(post.keywords)}
        matches += [typeahead_entry]
    return json.dumps(matches)


@blueprint.route('/ajax/index/typeahead_tags')
@blueprint.route('/ajax_tags_typeahead', methods=['GET'])
def generate_tags_typeahead():
    if not permissions.index_view.can():
        return '[]'
    return json.dumps([t[0] for t in db_session.query(Tag.name).all()])


@blueprint.route('/ajax/index/typeahead_users')
@blueprint.route('/ajax_users_typeahead', methods=['GET'])
def generate_users_typeahead():
    if not permissions.index_view.can():
        return '[]'
    return json.dumps([u[0] for u in db_session.query(User.identifier).all()])


@blueprint.route('/ajax/index/typeahead_paths')
@blueprint.route('/ajax_paths_typeahead', methods=['GET'])
def generate_projects_typeahead():
    if not permissions.index_view.can():
        return '[]'
    # return path stubs for all repositories
    stubs = [u'/'.join(p.split('/')[:-1]) for p in current_repo.dir()]
    return json.dumps(list(set(stubs)))



from __future__ import print_function

import cooked_input as ci
import logging
import os
import re
import time

import webbrowser

from .docx import DocxConverter

logger = logging.getLogger(__name__)


class GDocConverter(DocxConverter):
    _registry_keys = ['gdoc']

    def _find_doc(self, path, after=None, max_attempts=60, delay=1):
        """
        Look in the nominated path for a new "docx" document for a file modified
        after `after`. If none are found, retry every `delay` seconds for at
        most `max_attempts` attempts. If no documents are found, raise a
        `RuntimeError`.
        """
        count = 0
        while count < max_attempts:
            count += 1
            if count == 10:
                logger.info(
                    "So far unable to find a new 'docx' file in %s. If you "
                    "downloaded the Google Document elsewhere, please move it "
                    "into this folder.",
                    path
                )
            if not os.path.exists(path):
                break
            for filename in os.listdir(path):
                if filename.endswith('.docx'):
                    fpath = os.path.join(path, filename)
                    if os.path.getmtime(fpath) > after:
                        return fpath
            time.sleep(delay)

        if self.interactive:
            fpath = ci.get_string(
                prompt=(
                    "We were unable to find the downloaded Google Doc in the "
                    "expected path: '{}'. If the document was downloaded "
                    "elsewhere, please enter the full path of downloaded "
                    "document now (including file name)"
                    .format(path)
                ),
                required=False
            )
            if fpath:
                return fpath
        raise RuntimeError("Cannot find 'docx' document in {}.".format(path))

    def from_file(self, url, download_path=None, **opts):
        m = re.match('https://docs.google.com/document/d/(?P<doc_id>[^/]+)/', url)

        if not m:
            raise ValueError("Invalid Google Docs url.")

        doc_id = m.group('doc_id')
        download_url = "https://docs.google.com/document/d/{doc_id}/export?format=doc".format(doc_id=doc_id)

        time_start = time.time()
        webbrowser.open(download_url)

        time.sleep(2)

        download_path = download_path or os.path.expanduser('~/Downloads')
        logger.info("Looking for downloaded Google Docs file in '{}'...".format(download_path))
        filename = self._find_doc(download_path, after=time_start)

        DocxConverter.from_file(self, filename, **opts)

        headers = self.kp.headers
        if 'title' in headers and headers['title'].startswith('[]'):
            headers['title'] = re.sub(r'\[\]\{[^\}]+\}', '', headers['title'])
        if 'subtitle' in headers and headers['subtitle'].startswith('[]'):
            headers['subtitle'] = re.sub(r'\[\]\{[^\}]+\}', '', headers['subtitle'])
        self.kp.update_headers(**headers)



import os
import subprocess
import logging
import shutil

from .common import KnowledgeDeployer

logger = logging.getLogger()


class uWSGIDeployer(KnowledgeDeployer):

    _registry_keys = ['uwsgi']

    COMMAND = u"uwsgi {protocol} --plugin python --wsgi-file {{path}} --callable app --master --processes {{processes}} --threads {{threads}}"

    def run(self):
        if not self.app.check_thread_support():
            raise RuntimeError("Database configuration is not suitable for deployment (not thread-safe).")

        tmp_dir = self.write_temp_files()

        options = {
            'socket': u'{}:{}'.format(self.host, self.port),
            'processes': self.workers,
            'threads': 2,
            'timeout': self.timeout,
            'path': os.path.join(tmp_dir, 'server.py'),
            'cert': self.app.config['SSL_CERT']['cert'],
            'key': self.app.config['SSL_CERT']['key']
        }

        if self.app.config['DEPLOY_HTTPS']:
            self.COMMAND = self.COMMAND.format(protocol="--https {socket},{cert},{key}")
        else:
            self.COMMAND = self.COMMAND.format(protocol="--http {socket}")

        try:
            cmd = u"cd {};".format(tmp_dir) + self.COMMAND.format(**options)
            logger.info("Starting server with command: %s", u" ".join(cmd))
            subprocess.check_output(cmd, shell=True)
        finally:
            shutil.rmtree(tmp_dir)



import zipfile
import io
import os

from ..converter import KnowledgePostConverter


class PkpConverter(KnowledgePostConverter):
    _registry_keys = ['kp', 'zip']

    def to_file(self, filename):
        zf = zipfile.ZipFile(filename, 'w')

        for ref in self.kp._dir():
            zf.writestr(ref, self.kp._read_ref(ref))

        zf.close()

    def to_string(self):
        data = io.BytesIO()
        zf = zipfile.ZipFile(data, 'w')

        for ref in self.kp._dir():
            zf.writestr(ref, self.kp._read_ref(ref))

        zf.close()
        data.seek(0)
        return data.read()

    def from_file(self, filename):
        # Note: header checks are not applied here, since it should not be
        # possible to create portable knowledge post with incorrect headers.
        zf = zipfile.ZipFile(filename, 'r')

        for ref in zf.namelist():
            with zf.open(ref) as f:
                self.kp._write_ref(ref, f.read())

        zf.close()



from builtins import object

import os
from functools import wraps

from .postprocessor import KnowledgePostProcessor
from .utils.registry import SubclassRegisteringABCMeta
from .utils.dependencies import check_dependencies
from future.utils import with_metaclass


def get_format(filename, format=None):
    if format is None:
        if filename.startswith('https://docs.google.com/document/d/'):
            format = 'gdoc'
        elif filename.startswith('http://') or filename.startswith('https://'):
            format = 'proxy'
        elif '.' in filename:
            format = os.path.splitext(filename)[1]
            if format.startswith('.'):
                format = format[1:]
        else:
            raise RuntimeError("Unable to determine a format automatically. Please manually specify the format, and try again.")
    return format


class KnowledgePostConverter(with_metaclass(SubclassRegisteringABCMeta, object)):
    _registry_keys = None  # File extensions

    def __init__(self, kp, format=None, postprocessors=None, interactive=False, **kwargs):
        check_dependencies(self.dependencies, "Whoops! You are missing some dependencies required to use `{}` instances.".format(self.__class__.__name__))
        self.kp = kp
        self.format = format
        if postprocessors is None:
            postprocessors = [('extract_images', {}), ('format_checks', {})]
        self.postprocessors = postprocessors
        self.interactive = interactive
        self.init(**kwargs)

    @property
    def dependencies(self):
        return []

    def init(self):
        pass

    def __get_wrapped_with_postprocessors(self, f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            try:
                kp = self.kp
                f(*args, **kwargs)
                postprocessors = self.postprocessors

                if f.__name__ == 'from_file':
                    filename = args[0] if len(args) > 0 else kwargs['filename']
                    kp.orig_context = os.path.dirname(filename)

                if postprocessors is None:
                    postprocessors = []
                for postprocessor, kwargs in postprocessors:
                    KnowledgePostProcessor._get_subclass_for(postprocessor)(**kwargs).process(kp)

                return kp
            finally:
                self.cleanup()
        return wrapped

    def __getattribute__(self, attr):
        if attr in ['from_file', 'from_string']:
            return self.__get_wrapped_with_postprocessors(object.__getattribute__(self, attr))
        return object.__getattribute__(self, attr)

    def kp_write(self, md, headers=None, images={}):
        return self.kp.write(md, headers=headers, images=images, interactive=self.interactive)

    def from_file(self, filename, **opts):
        raise NotImplementedError

    def from_string(self, string, **opts):
        raise NotImplementedError

    def to_file(self, filename, **opts):
        raise NotImplementedError

    def to_string(self, **opts):
        raise NotImplementedError

    def cleanup(self):
        pass

    @classmethod
    def for_file(cls, kp, filename, format=None, postprocessors=None, interactive=False):
        return cls.for_format(kp, get_format(filename, format), postprocessors=postprocessors, interactive=interactive)

    @classmethod
    def for_format(cls, kp, format, postprocessors=None, interactive=False):
        if format.lower() not in cls._registry:
            raise ValueError("The knowledge repository does not support files of type '{}'. Supported types are: {}."
                             .format(format, ','.join(list(cls._registry.keys()))))
        return cls._get_subclass_for(format.lower())(kp, format=format, postprocessors=postprocessors, interactive=interactive)



from . import auth
from . import health
from . import comment
from . import index
from . import posts
from . import stats
from . import tags
from . import vote
from . import editor
from . import groups
from . import debug



import os
import shutil
import random
import string
import logging
import time
import tempfile

try:
    from urllib.parse import urljoin
except ImportError:  # Python 2
    from urlparse import urljoin

from .extract_images import ExtractImages


logger = logging.getLogger(__name__)


class ExtractImagesToLocalServer(ExtractImages):
    """
    This KnowledgePostProcessor subclass extracts images from posts to a local
    directory. It is assumed that a local http server is then serving these images
    from the local directory. It is designed to be used upon addition to a knowledge
    repository, which can reduce the size of repositories. It replaces local images
    with urls relative to `http_image_root` (the base url of the local http server).

    `image_dir` should be the root of the image folder which is accessible locally.
    `http_image_root` should be the root of the server where the images will be
    accessible (e.g. 'http://localhost:8000').
    """

    _registry_keys = ['extract_images_to_local']

    def __init__(self, image_dir, http_image_root):
        self.image_dir = image_dir
        self.http_image_root = http_image_root

    def copy_image(self, kp, img_path, is_ref=False, repo_name='knowledge'):
        # Copy image data to new file
        if is_ref:
            _, tmp_path = tempfile.mkstemp()
            with open(tmp_path, 'wb') as f:
                f.write(kp._read_ref(img_path))
        else:
            tmp_path = img_path

        try:
            # Get image type
            img_ext = os.path.splitext(img_path)[1]

            # Make random filename for image
            random_string = ''.join(
                random.choice(string.ascii_lowercase) for i in range(6))
            fname_img = '{repo_name}_{time}_{random_string}{ext}'.format(
                repo_name=repo_name,
                time=int(round(time.time() * 100)),
                random_string=random_string,
                ext=img_ext).strip().replace(' ', '-')

            # See if a static file directory exists, if not, let's create
            if not os.path.exists(self.image_dir):
                os.makedirs(self.image_dir)

            # Copy images to local http server directory
            new_path = os.path.join(self.image_dir, fname_img)
            logger.info("Copying image {} to {}".format(tmp_path, new_path))
            try:
                shutil.copyfile(tmp_path, new_path)
            except Exception as e:
                raise Exception('Problem copying image: {}'.format(e))

        finally:
            # Clean up temporary file
            if is_ref:
                os.remove(tmp_path)

        # return uploaded path of file
        return urljoin(self.http_image_root, fname_img)

    def skip_image(self, kp, image):
        import re
        if re.match('http[s]?://', image['src']):
            return True
        return False

    def cleanup(self, kp):
        if kp._has_ref('images'):
            kp._drop_ref('images')



import datetime

from future.moves.urllib.parse import urlparse, urlencode, urljoin

from flask import request, url_for
from flask_login import AnonymousUserMixin, login_user
from flask_principal import Identity, identity_changed, UserNeed

from .. import roles
from ..proxies import current_app, db_session


def prepare_user(user, session_start=True):
    cache_lifetime = current_app.config['AUTH_USER_ATTRIBUTE_CACHE_LIFETIME'] or 0
    if (
        current_app.config['AUTH_USER_ATTRIBUTE_SETTER'] and
        (
            session_start or
            user.last_login_at is None or
            user.last_login_at < datetime.datetime.now() - datetime.timedelta(seconds=cache_lifetime)
        )
    ):
        session_start = True
        user = current_app.config['AUTH_USER_ATTRIBUTE_SETTER'](user)

    if session_start or user.id is None:
        user.last_login_at = datetime.datetime.now()
        db_session.add(user)
        db_session.commit()

    return user


class AnonymousKnowledgeUser(AnonymousUserMixin):
    id = None
    identifier = None
    created_at = None
    posts = []

    @property
    def subscriptions(self):
        return []

    @property
    def liked_posts(self):
        return []


def is_safe_url(target):
    ref_url = urlparse(request.host_url)
    test_url = urlparse(urljoin(request.host_url, target))
    return test_url.scheme in ('http', 'https') and ref_url.netloc == test_url.netloc


def populate_identity_roles(identity, user=None):
    identity.user = user

    if user is None or user.is_anonymous:
        if current_app.config['POLICY_ANONYMOUS_VIEW_INDEX']:
            identity.provides.add(roles.index_view)
        if current_app.config['POLICY_ANONYMOUS_VIEW_POST']:
            identity.provides.add(roles.post_view)
        if current_app.config['POLICY_ANONYMOUS_VIEW_STATS']:
            identity.provides.add(roles.stats_view)
        if current_app.config['POLICY_ANONYMOUS_DOWNLOADS']:
            identity.provides.add(roles.post_download)

    else:
        identity.provides.add(UserNeed(user.identifier))
        identity.provides.add(roles.index_view)
        identity.provides.add(roles.post_view)
        identity.provides.add(roles.post_edit)
        identity.provides.add(roles.post_comment)
        identity.provides.add(roles.post_download)
        identity.provides.add(roles.stats_view)

        # TODO: Populate group permissions, and port existing group admin
        # code to roles.

    return identity



import logging
import os
from builtins import str
from flask import request, url_for, redirect, render_template, current_app, Blueprint, g, Response, abort

from .. import permissions
from ..proxies import db_session, current_repo, current_user
from ..models import User, Post, PageView
from ..utils.render import render_post, render_comment, render_post_raw


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


blueprint = Blueprint('posts', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/post/<path:path>', methods=['GET'])
@PageView.logged
@permissions.post_view.require()
def render(path):
    """
    Render the knowledge post with all the related formatting.
    """

    mode = request.args.get('render', 'html')
    username, user_id = current_user.identifier, current_user.id

    tmpl = 'markdown-rendered.html'
    if mode == 'raw':
        tmpl = 'markdown-raw.html'
    elif mode == 'presentation':
        # TODO(dan?) fix presentation post
        # presentation_post = {}
        # presentation_post['authors_string'] = post.author_string
        # presentation_post['tldr'] = post.tldr
        # presentation_post['html'] = html
        # html = create_presentation_text(presentation_post)
        tmpl = "markdown-presentation.html"

    if not current_app.config.get('INDEXING_ENABLED', True):
        return _render_preview(path=path, tmpl=tmpl)

    post = (db_session.query(Post)
                      .filter(Post.path == path)
                      .first())
    if not post:
        knowledge_aliases = current_repo.config.aliases
        if path in knowledge_aliases:
            # TODO: reframe as redirect
            post = (db_session.query(Post)
                              .filter(Post.path == knowledge_aliases[path])
                              .first())
    if not post:
        logger.warning(u"unable to find post at {}".format(path))
        return abort(404)
    if post.contains_excluded_tag:
        # It's possible that someone gets a direct link to a post that has an excluded tag
        return render_template("error.html")

    if post.private and not (username in post.authors or username in current_repo.config.editors):
        allowed_users = set(user.id for group in post.groups for user in group.users)
        if user_id not in allowed_users:
            return render_template("permission_ask.html", authors=post.authors_string)

    rendered = render_post(post, with_toc=True)
    raw_post = render_post_raw(post) if mode == 'raw' else None

    comments = post.comments
    for comment in comments:
        author = db_session.query(User).filter(User.id == comment.user_id).first()
        if author is not None:
            comment.author = author.format_name
        else:
            comment.author = 'Anonymous'
        if mode != 'raw':
            comment.text = render_comment(comment)

    user_obj = current_user

    tags_list = [str(t.name) for t in post.tags]
    user_subscriptions = [str(s) for s in user_obj.subscriptions]

    is_author = user_id in [author.id for author in post.authors]

    web_editor_prefixes = current_app.config['WEB_EDITOR_PREFIXES']
    is_webpost = False
    if web_editor_prefixes:
        is_webpost = any(prefix for prefix in web_editor_prefixes if path.startswith(prefix))

    rendered = render_template(tmpl,
                               html=rendered['html'],
                               toc=rendered['toc'],
                               post_id=post.id,
                               post_path=path,
                               raw_post=raw_post,
                               comments=comments,
                               username=username,
                               post_author=post.authors_string,
                               title=post.title,
                               page_views=post.view_count,
                               unique_views=post.view_user_count,
                               likes=post.vote_counted_for_user(user_id=user_id),
                               total_likes=post.vote_count,
                               tags_list=tags_list,
                               user_subscriptions=user_subscriptions,
                               show_webeditor_button=is_webpost and is_author,
                               webeditor_buttons=is_webpost,
                               web_uri=post.kp.web_uri,
                               table_id=None,
                               is_private=(post.private == 1),
                               is_author=is_author,
                               can_download=permissions.post_download.can(),
                               downloads=post.kp.src_paths)
    return rendered


@render.object_extractor
def render(path):
    return {
        'id': Post.query.filter(Post.path == path).first().id,
        'type': 'post',
        'action': 'view'
    }


@blueprint.route('/post/preview/<path:path>', methods=['GET'])
@PageView.logged
@permissions.post_view.require()
def render_preview(path):
    return _render_preview(path, 'markdown-rendered.html')


def _render_preview(path, tmpl):
    post = None
    mode = request.args.get('render', 'html')

    if current_repo.has_post(path):
        post = current_repo.post(path)
    else:
        knowledge_aliases = current_repo.config.aliases
        if path in knowledge_aliases:
            # TODO: reframe as redirect
            post = current_repo.post(knowledge_aliases[path])

    if not post:
        raise Exception(u"unable to find post at {}".format(path))

    rendered = render_post(post, with_toc=True)
    raw_post = render_post_raw(post) if (mode == 'raw') else None

    return render_template(tmpl,
                           html=rendered['html'],
                           toc=rendered['toc'],
                           post_id=None,
                           post_path=path,
                           raw_post=raw_post,
                           comments=[],
                           username=None,
                           post_author=u', '.join(post.headers['authors']),
                           title=post.headers['title'],
                           page_views=0,
                           unique_views=0,
                           likes=1,
                           total_likes=0,
                           tags_list=post.headers.get('tags', []),
                           user_subscriptions=[],
                           webeditor_buttons=False,
                           table_id=None)


# DEPRECATED: Legacy route for the /render endpoint to allow old bookmarks to function
@blueprint.route('/render', methods=['GET'])
@PageView.logged
@permissions.post_view.require()
def render_legacy():
    path = request.args.get('markdown', '')
    return redirect(url_for('.render', path=path), code=302)


@blueprint.route('/ajax/post/download', methods=['GET'])
@PageView.logged
@permissions.post_download.require()
def download():
    "Downloads resources associated with a post."

    post = current_repo.post(request.args.get('post'))  # Note: `post` is expected to be a post path
    resource_type = request.args.get('type', 'source')

    if resource_type in ('kp', 'zip'):
        filename = os.path.basename(post.path)
        if resource_type == 'zip':
            filename = filename[:-3] + '.zip'
        return Response(
            post.to_string(format=resource_type),
            mimetype="application/zip",
            headers={u"Content-disposition": "attachment; filename={}".format(filename)})
    elif resource_type == 'source':
        path = request.args.get('path', None)
        assert path is not None, "Source path not provided."
        assert path in post.src_paths, "Provided reference is not a valid source path."
        return Response(
            post._read_ref(path),
            mimetype="application/octet-stream",
            headers={u"Content-disposition": "attachment; filename={}".format(os.path.basename(path))})
    else:
        raise RuntimeError(u"Invalid resource_type: {}".format(resource_type))



import os
import posixpath
import re
import logging

from ..postprocessor import KnowledgePostProcessor

logger = logging.getLogger(__name__)


class ExtractImages(KnowledgePostProcessor):
    _registry_keys = ['extract_images']

    def process(self, kp):
        images = self.find_images(kp.read())
        image_mapping = self.collect_images(kp, images)
        self.update_thumbnail_uri(kp, images, image_mapping)
        self.cleanup(kp)

    def update_thumbnail_uri(self, kp, images, image_mapping):
        thumbnail = kp.headers.get('thumbnail', 0)

        # if thumbnail is a number, select the nth image in this post as the thumbnail
        if isinstance(thumbnail, str) and thumbnail.isdigit():
            thumbnail = int(thumbnail)
        if isinstance(thumbnail, int):
            if len(images) > 0:
                image_index = 0
                if thumbnail < len(images):
                    image_index = thumbnail
                thumbnail = images[image_index]['src']
            else:
                thumbnail = None

        # if thumbnail is a url, copy it locally to the post unless already collected during collection
        if thumbnail and not self.skip_image(kp, {'src': thumbnail}):
            orig_path = os.path.join(kp.orig_context, os.path.expanduser(thumbnail))
            if thumbnail in image_mapping:
                thumbnail = image_mapping[thumbnail]
            elif os.path.exists(orig_path):
                thumbnail = self.copy_image(kp, orig_path)
            else:
                logger.warning("Could not find a thumbnail image at: {}".format(thumbnail))

        # update post headers to point to new thumbnail image
        kp.update_headers(thumbnail=thumbnail)

    def find_images(self, md):
        images = []
        images.extend(self.collect_images_for_pattern(
            md,
            # Match all <img /> tags with attributes of form <name>(="<value>") with optional quotes (can be single or double)
            # The src attribute is exected to always be surrounded by quotes.
            r'<img\s+(?:\w+(?:=([\'\"])?(?(1)(?:(?!\1).)*?\1|[^>]*?))?\s+?)*src=([\'\"])(?P<src>(?:(?!\2).)*?)\2(?:\s+\w+(?:=([\'\"])?(?(1)(?:(?!\4).)*?\4|[^>]*?))?)*\s*\/?>'
        ))
        images.extend(self.collect_images_for_pattern(md, r'\!\[[\s\S]*?\]\((?P<src>[^\)]*)\)'))
        return sorted(images, key=lambda x: x['offset'])

    def collect_images_for_pattern(self, md, pattern=None):
        p = re.compile(pattern)
        return [{'offset': m.start(), 'tag': m.group(0), 'src': m.group('src')} for m in p.finditer(md)]

    def collect_images(self, kp, images):
        image_mapping = {}
        if len(images) == 0:
            return image_mapping
        md = kp.read()
        images = images[::-1]
        for image in images:
            if self.skip_image(kp, image):
                continue
            orig_path = os.path.join(kp.orig_context, os.path.expanduser(image['src']))
            new_path = None
            if image['src'] in image_mapping:
                new_path = image_mapping[image['src']]
            elif kp._has_ref(image['src']):
                new_path = self.copy_image(kp, image['src'], is_ref=True)
            elif os.path.exists(orig_path):
                new_path = self.copy_image(kp, orig_path)
            else:
                logger.warning("Could not find an image at: {}".format(image['src']))
            if not new_path:
                continue
            image_mapping[image['src']] = new_path
            md = self.replace_image_locations(md, image['offset'], image['tag'], image['src'], new_path)
        kp.write(md)
        return image_mapping

    def skip_image(self, kp, image):
        if re.match('http[s]?://', image['src']):
            return True
        if image['src'].startswith('images/') and image['src'] in kp.image_paths:
            return True
        return False

    def copy_image(self, kp, path, is_ref=False):
        if is_ref:
            return
        with open(path, 'rb') as f:
            kp.write_image(os.path.basename(path), f.read())
        return posixpath.join('images', os.path.basename(path))

    def replace_image_locations(self, md, offset, match, old_path, new_path):
        pre = md[:offset]
        post = md[offset + len(match):]
        return pre + match.replace(old_path, new_path) + post

    def cleanup(self, kp):
        pass



import os
import posixpath
import random
import string
import logging
import tempfile
import time

from .extract_images import ExtractImages


logger = logging.getLogger(__name__)


class ExtractImagesToS3(ExtractImages):
    '''
    This KnowledgePostProcessor subclass extracts images from posts to S3. It
    is designed to be used upon addition to a knowledge repository, which can
    reduce the size of repositories. It replaces local images with remote urls
    based on `http_image_root`.

    `s3_image_root` should be the root of the image folder on an S3 remote, such
    as "s3://my_bucket/images".
    `http_image_root` should be the root of the server where the images will be
    accessible after uploading.

    Note: This requires that user AWS credentials are set up appropriately and
    that they have installed the aws cli packages.
    '''
    _registry_keys = ['extract_images_to_s3']

    def __init__(self, s3_image_root, http_image_root):
        self.s3_image_root = s3_image_root
        self.http_image_root = http_image_root

    def copy_image(self, kp, img_path, is_ref=False, repo_name='knowledge'):
        # Copy image data to new file
        if is_ref:
            _, tmp_path = tempfile.mkstemp()
            with open(tmp_path, 'wb') as f:
                f.write(kp._read_ref(img_path))
        else:
            tmp_path = img_path

        try:
            # Get image type
            img_ext = posixpath.splitext(img_path)[1]

            # Make random filename for image
            random_string = ''.join(random.choice(string.ascii_lowercase) for i in range(6))
            fname_img = '{repo_name}_{time}_{random_string}{ext}'.format(
                repo_name=repo_name,
                time=int(round(time.time() * 100)),
                random_string=random_string,
                ext=img_ext).strip().replace(' ', '-')

            # Copy image to accessible folder on S3
            fname_s3 = posixpath.join(self.s3_image_root, repo_name, fname_img)
            # Note: The following command may need to be prefixed with a login agent;
            # for example, to handle multi-factor authentication.
            cmd = "aws s3 cp '{0}' {1}".format(tmp_path, fname_s3)
            logger.info("Uploading images to S3: {cmd}".format(cmd=cmd))
            retval = os.system(cmd)
            if retval != 0:
                raise Exception('Problem uploading images to s3')
        finally:
            # Clean up temporary file
            if is_ref:
                os.remove(tmp_path)

        # return uploaded path of file
        return posixpath.join(self.http_image_root, repo_name, fname_img)

    def skip_image(self, kp, image):
        import re
        if re.match('http[s]?://', image['src']):
            return True
        return False

    def cleanup(self, kp):
        if kp._has_ref('images'):
            kp._drop_ref('images')



from functools import partial
from flask_principal import RoleNeed, ItemNeed

# Need types
GroupNeed = partial(ItemNeed, 'group')

# Site roles
admin = RoleNeed('admin')
stats_view = RoleNeed('stats_view')

# Index roles
index_view = RoleNeed('index_view')

# Post roles
post_comment = RoleNeed('post_comment')
post_edit = RoleNeed('post_edit')
post_view = RoleNeed('post_view')
post_download = RoleNeed('post_download')



import datetime

import six
from past.builtins import basestring

from ..post import HEADER_OPTIONAL_FIELD_TYPES, HEADER_REQUIRED_FIELD_TYPES
from ..postprocessor import KnowledgePostProcessor


class FormatChecks(KnowledgePostProcessor):
    _registry_keys = ['format_checks']

    def process(self, kp):
        headers = kp.headers
        for field, typ, input in HEADER_REQUIRED_FIELD_TYPES:
            assert field in headers, "Required field `{}` missing from headers.".format(
                field)
            assert isinstance(headers[field], typ), "Value for field `{}` is of type {}, and needs to be of type {}.".format(
                field, type(headers[field]), typ)
        for field, typ, input in HEADER_OPTIONAL_FIELD_TYPES:
            if field in headers:
                assert isinstance(headers[field], typ), "Value for field `{}` is of type {}, and needs to be of type {}.".format(
                    field, type(headers[field]), typ)



import json
import logging
import sys
import os
from builtins import str
from datetime import datetime
from flask import request, render_template, Blueprint, current_app, url_for, send_from_directory, g
from sqlalchemy import or_
from werkzeug import secure_filename

from knowledge_repo.post import KnowledgePost
from .. import permissions
from ..proxies import db_session, current_repo, current_user
from ..models import Post, PostAuthorAssoc, Tag, Comment, User, PageView
from ..utils.emails import send_review_email, send_reviewer_request_email
from ..utils.image import pdf_page_to_png, is_pdf, is_allowed_image_format

from ..index import update_index

if sys.version_info.major > 2:
    from urllib.parse import unquote as urlunquote
else:
    from urllib import unquote as urlunquote

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


blueprint = Blueprint('editor', __name__,
                      template_folder='../templates', static_folder='../static')

# TODO: These functions have not been fully married to the KnowledgePost API
# Currently, backended by Post objects but partially implemented on KnowledgePost API


# TODO: Deprecate this route in favour of integrating editing links into primary index pages and user pages
@blueprint.route('/webposts', methods=['GET'])
@PageView.logged
@permissions.post_edit.require()
def gitless_drafts():
    """ Render the gitless posts that a user has created in table form
        Editors can see all the posts created via Gitless_Editing
    """
    prefixes = current_app.config.get('WEB_EDITOR_PREFIXES', [])
    if prefixes == []:
        raise Exception("Web editing is not configured")

    query = (db_session.query(Post))
    if prefixes is not None:
        query = query.filter(or_(*[Post.path.like(p + '%') for p in prefixes]))

    if current_user.identifier not in current_repo.config.editors:
        query = (query.outerjoin(PostAuthorAssoc, PostAuthorAssoc.post_id == Post.id)
                      .filter(PostAuthorAssoc.user_id == current_user.id))

    return render_template("web_posts.html", posts=query.all())


@blueprint.route('/edit')
@blueprint.route('/edit/<path:path>', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def editor(path=None):
    """ Render the web post editor, either with the default values
        or if the post already exists, with what has been saved """

    prefixes = current_app.config.get('WEB_EDITOR_PREFIXES', None)

    if prefixes is not None:
        assert (
            path is None or any(path.startswith(prefix) for prefix in prefixes)
        ), "Editing of this post online is not permitted by server configuration."

    # set defaults
    data = {'title': None,
            'status': current_repo.PostStatus.DRAFT.value,
            'markdown': request.args.get('markdown'),
            'thumbnail': '',
            'can_approve': 0,
            'username': current_user.identifier,
            'created_at': datetime.now(),
            'updated_at': datetime.now(),
            'authors': [current_user.identifier],
            'comments': [],
            'tldr': request.args.get('tldr'),
            }

    if path is not None and path in current_repo:
        kp = current_repo.post(path)
        data.update(kp.headers)

        data['status'] = kp.status.value
        data['path'] = path
        data['markdown'] = kp.read(images=False, headers=False)

        # retrieve reviews
        post = db_session.query(Post).filter(Post.path == path).first()
        if post:  # post may have not been indexed yet
            data['comments'] = (db_session.query(Comment)
                                          .filter(Comment.post_id == post.id)
                                          .filter(Comment.type == "review")
                                          .all())

    if current_user.identifier not in data['authors'] or current_user.identifier in current_repo.config.editors:
        data['can_approve'] = 1

    data['created_at'] = data['created_at']
    data['updated_at'] = data['updated_at']
    data['authors'] = json.dumps(data.get('authors'))
    data['tags'] = json.dumps(data.get('tags', []))

    if 'proxy' in data or request.args.get('proxy', False):
        return render_template('post_editor_proxy.html', **data)
    return render_template('post_editor_markdown.html',
                           **data)


@blueprint.route('/ajax/editor/save', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def save_post():
    """ Save the post """

    data = request.get_json()
    path = data['path']

    prefixes = current_app.config['WEB_EDITOR_PREFIXES']
    if prefixes == []:
        raise Exception("Web editing is not configured")

    if prefixes is not None:
        if not any([path.startswith(prefix) for prefix in prefixes]):
            return json.dumps({'msg': (u"Your post path must begin with one of {}").format(prefixes),
                               'success': False})

    # TODO better handling of overwriting
    kp = None
    if path in current_repo:
        kp = current_repo.post(path)
        if current_user.identifier not in kp.headers['authors'] and current_user.identifier not in current_repo.config.editors:
            return json.dumps({'msg': (u"Post with path {} already exists and you are not an author!"
                                       "\nPlease try a different path").format(path),
                               'success': False})

    # create the knowledge post
    kp = kp or KnowledgePost(path=path)

    headers = {}
    headers['created_at'] = datetime.strptime(data['created_at'], '%Y-%m-%d').date()
    headers['updated_at'] = datetime.strptime(data['updated_at'], '%Y-%m-%d').date()
    headers['title'] = data['title']
    headers['path'] = data['path']
    # TODO: thumbnail header not working currently, as feed image set with kp
    # method not based on header
    headers['thumbnail'] = data.get('feed_image', '')
    headers['authors'] = [auth.strip() for auth in data['author']]
    headers['tldr'] = data['tldr']
    headers['tags'] = [tag.strip() for tag in data.get('tags', [])]
    if 'proxy' in data:
        headers['proxy'] = data['proxy']

    kp.write(urlunquote(data['markdown']), headers=headers)
    # add to repo
    current_repo.add(kp, update=True, message=headers['title'])  # THIS IS DANGEROUS

    update_index()
    return json.dumps({'path': path})


@blueprint.route('/ajax/editor/submit', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def submit_for_review():
    """ Submit post and if there are reviewers assigned, email them"""
    path = request.args.get('path', None)
    data = request.get_json()
    current_repo.submit(path)

    # email the reviewers
    reviewers = data.get('post_reviewers', None)
    if reviewers:
        for r in reviewers.split(','):
            send_reviewer_request_email(path=path, reviewer=r)

    update_index()
    return 'OK'


@blueprint.route('/ajax/editor/publish', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def publish_post():
    """ Publish the post by changing the status """
    path = request.args.get('path', None)
    if path not in current_repo:
        return json.dumps({'msg': u"Unable to retrieve post with path = {}!".format(path), 'success': False})
    current_repo.publish(path)

    update_index(check_timeouts=False)
    return 'OK'


@blueprint.route('/ajax/editor/unpublish', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def unpublish_post():
    """ Unpublish the post """
    path = request.args.get('path', None)
    if path not in current_repo:
        return json.dumps({'msg': u"Unable to retrieve post with path = {}!".format(path), 'success': False})
    current_repo.unpublish(path)

    update_index(check_timeouts=False)
    return 'OK'


@blueprint.route('/ajax/editor/accept', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def accept():
    """ Accept the post """
    path = request.args.get('path', None)
    if path not in current_repo:
        return json.dumps({'msg': u"Unable to retrieve post with path = {}!".format(path), 'success': False})
    current_repo.accept(path)
    update_index()
    return 'OK'


@blueprint.route('/ajax/editor/delete', methods=['GET', 'POST'])
@PageView.logged
@permissions.post_edit.require()
def delete_post():
    """ Delete a post """
    path = request.args.get('path', None)
    if path not in current_repo:
        return json.dumps({'msg': u"Unable to retrieve post with path = {}!".format(path), 'success': False})
    kp = current_repo.post(path)
    if current_user.identifier not in kp.headers['authors']:
        return json.dumps({'msg': "You can only delete a post where you are an author!", 'success': False})
    current_repo.remove(path)

    update_index(check_timeouts=False)
    return 'OK'


@blueprint.route('/ajax/editor/review', methods=['POST', 'DELETE'])
@PageView.logged
@permissions.post_edit.require()
def review_comment():
    """
    Saves a review and sends an email that the post has been reviewed to the author of the post or deletes a submitted review
    """

    if request.method == 'POST':
        path = request.args.get('path', None)
        post_id = db_session.query(Post).filter(Post.path == path).first().id

        comment = Comment()
        comment.text = request.get_json()['text']
        comment.user_id = current_user.id
        comment.post_id = post_id
        comment.type = "review"
        db_session.add(comment)
        db_session.commit()

        send_review_email(path=path,
                          commenter=current_user.identifier,
                          comment_text=comment.text)

    elif request.method == 'DELETE':
        comment = Comment.query.get(int(request.args.get('comment_id', '')))
        if comment and current_user.id == comment.user_id:
            db_session.delete(comment)
            db_session.commit()

    return 'OK'


# DEPRECATED
@blueprint.route('/file_upload', methods=['POST', 'GET'])
@PageView.logged
@permissions.post_edit.require()
def file_upload():
    """ Uploads images dropped on the web editor's markdown box to static/images
        and notifies editors by email
    """
    upload_folder = 'images'
    title = request.form['title']
    files = request.files
    uploadedFiles = []

    if files:
        for img_file in files.values():
            filename = secure_filename(title + "_" + img_file.filename).lower()
            dst_folder = os.path.join(current_app.static_folder, upload_folder)

            if is_allowed_image_format(img_file):
                try:
                    img_file.save(os.path.join(dst_folder, filename))
                    send_from_directory(dst_folder, filename)
                    uploadedFiles += [url_for("static", filename=os.path.join(upload_folder, filename))]
                except Exception as e:
                    error_msg = u"ERROR during image upload: {}".format(str(e))
                    logger.error(error_msg)
                    return json.dumps({'error_msg': error_msg, 'success': False})

            elif is_pdf(filename):
                from PyPDF2 import PdfFileReader
                try:
                    src_pdf = PdfFileReader(img_file)
                    filename = os.path.splitext(filename)[0]
                    num_pages = src_pdf.getNumPages()
                    for page_num in range(num_pages):
                        page_png = pdf_page_to_png(src_pdf, page_num)
                        page_name = u"{filename}_{page_num}.jpg".format(**locals())
                        page_png.save(filename=os.path.join(dst_folder, page_name))
                        uploadedFiles += [url_for("static", filename=os.path.join(upload_folder, page_name))]
                except Exception as e:
                    error_msg = u"ERROR during pdf upload: {}".format(str(e))
                    logger.error(error_msg)
                    return json.dumps({'error_msg': error_msg, 'success': False})

    return json.dumps({'links': uploadedFiles, 'success': True})



import re

from ..converter import KnowledgePostConverter


class ProxyConverter(KnowledgePostConverter):

    _registry_keys = ['proxy']

    def from_file(self, url, **opts):

        # Deal with special cases, whereby url should be mutated before being
        # added to post headers.

        # Google presentations should be embedded in "embed" mode.
        gpres = re.match('^https://docs.google.com/presentation/d/(?P<pres_id>[^/]+)/edit(?P<slide_query>.*)$', url)
        if gpres:
            url = "https://docs.google.com/presentation/d/{}/embed{}".format(*gpres.groups())

        self.kp_write("", headers={'proxy': url})



from __future__ import absolute_import

import multiprocessing
import time
import atexit
import os
import imp
import logging
import traceback
import math
import multiprocessing
import uuid
import mimetypes

from flask import Flask, current_app, render_template, g, request, flash, redirect, url_for, session
from flask_login import LoginManager, user_loaded_from_request
from flask_mail import Mail
from flask_migrate import Migrate
from flask_principal import Principal, identity_loaded, identity_changed, Identity, RoleNeed, UserNeed, AnonymousIdentity, PermissionDenied
from alembic import command
from alembic.migration import MigrationContext
from datetime import datetime
from werkzeug import url_encode

import knowledge_repo
from . import roles, routes
from .auth_provider import KnowledgeAuthProvider
from .proxies import db_session, current_repo, current_user
from .index import update_index, set_up_indexing_timers, time_since_index, time_since_index_check
from .models import db as sqlalchemy_db, Post, User, Tag
from .utils.auth import AnonymousKnowledgeUser, populate_identity_roles, prepare_user

# Needed to serve svg with correct mime type over https
mimetypes.add_type('image/svg+xml', '.svg')

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class KnowledgeFlask(Flask):

    def __init__(self, repo, db_uri=None, debug=None, config=None, **kwargs):
        Flask.__init__(self, __name__,
                       template_folder='templates',
                       static_folder='static')

        # Add unique identifier for this application isinstance
        self.uuid = str(uuid.uuid4())
        if 'KNOWLEDGE_REPO_MASTER_UUID' not in os.environ:
            logger.info("Set KNOWLEDGE_REPO_MASTER_UUID to '{}'.".format(self.uuid))
            os.environ['KNOWLEDGE_REPO_MASTER_UUID'] = self.uuid

        # Preload default configuration
        self.config.from_object('knowledge_repo.app.config_defaults')

        # Load configuration from file or provided object
        if config:
            if isinstance(config, str):
                config = imp.load_source('knowledge_server_config', os.path.abspath(config))
            self.config.from_object(config)

        # Add configuration passed in as keyword arguments
        self.config.update(kwargs)

        # Prepare repository, and add it to the app
        if hasattr(config, 'prepare_repo'):
            repo = config.prepare_repo(repo)
        self.repository = repo
        assert isinstance(self.repository, knowledge_repo.KnowledgeRepository), "Invalid repository object provided."

        # Set debug mode from kwargs or else maintain current setting
        if debug is not None:
            self.config['DEBUG'] = debug

        # Set the secret key for this instance (creating one if one does not exist already)
        self.config['SECRET_KEY'] = self.config['SECRET_KEY'] or str(uuid.uuid4())

        # Configure database
        if db_uri:
            self.config['SQLALCHEMY_DATABASE_URI'] = db_uri
        logger.debug(u"Using database: {}".format(self.config['SQLALCHEMY_DATABASE_URI']))

        # Register database schema with flask app
        sqlalchemy_db.init_app(self)

        # Set up database migration information
        # Registers Migrate plugin in self.extensions['migrate']
        Migrate(self, self.db)

        # Try to create the database if it does not already exist
        # Existence is determined by whether there is an existing alembic migration revision
        db_auto_create = self.config.get('DB_AUTO_CREATE', True)
        db_auto_upgrade = self.config.get('DB_AUTO_UPGRADE', True)
        if db_auto_create and self.db_revision is None:
            self.db_init()
        elif db_auto_upgrade:
            self.db_upgrade()

        # Ensure that indexes are set up if required by the time first page is
        # served.
        @self.before_first_request
        def start_indexing():
            if self.config['INDEXING_ENABLED']:
                self.start_indexing()

        # Initialise login manager to keep track of user sessions
        LoginManager().init_app(self)
        self.login_manager.login_view = 'auth.login'
        self.login_manager.anonymous_user = AnonymousKnowledgeUser

        @self.login_manager.user_loader
        def load_user(user_id):
            return User(identifier=user_id)

        # Attempt login via http headers
        if self.config.get('AUTH_USE_REQUEST_HEADERS'):
            @self.login_manager.request_loader
            def load_user_from_request(request):
                user_attributes = current_app.config.get('AUTH_MAP_REQUEST_HEADERS')(request.headers)
                if isinstance(user_attributes, dict) and user_attributes.get('identifier', None):
                    user = User(identifier=user_attributes['identifier'])
                    user.can_logout = False
                    for attribute in ['avatar_uri', 'email', 'name']:
                        if user_attributes.get(attribute):
                            setattr(user, attribute, user_attributes[attribute])
                    user = prepare_user(user, session_start=False)
                    return user
        elif self.config.get('AUTH_USER_IDENTIFIER_REQUEST_HEADER'):
            logger.warning("AUTH_USER_IDENTIFIER* configuration keys are deprecated and will be removed in v0.9.0 .")

            @self.login_manager.request_loader
            def load_user_from_request(request):
                identifier = request.headers.get(current_app.config['AUTH_USER_IDENTIFIER_REQUEST_HEADER'])
                if identifier:
                    if current_app.config['AUTH_USER_IDENTIFIER_REQUEST_HEADER_MAPPING']:
                        identifier = current_app.config['AUTH_USER_IDENTIFIER_REQUEST_HEADER_MAPPING'](identifier)
                    user = User(identifier=identifier)
                    user.can_logout = False
                    user = prepare_user(user, session_start=False)
                    return user

        # Intialise access policies
        self.principal = Principal(self)

        # Add AnonymousIdentity fallback so that on_identity_loaded is called for
        # anonymous users too.
        self.principal.identity_loaders.append(lambda: AnonymousIdentity())

        # Synchronise user permissions with data model
        @user_loaded_from_request.connect
        def on_user_loaded_from_request(sender, user):
            self.principal.set_identity(Identity(user.id))

        @identity_loaded.connect_via(self)
        def on_identity_loaded(sender, identity):
            populate_identity_roles(identity, user=current_user)

        @self.errorhandler(PermissionDenied)
        def handle_insufficient_permissions(error):
            session['requested_url'] = request.url
            return render_template('permission_denied.html'), 403

        # Add mail object if configuration is supplied
        if self.config.get('MAIL_SERVER'):
            self.config['mail'] = Mail(self)

        # Register routes to be served
        self.register_blueprint(routes.posts.blueprint)
        self.register_blueprint(routes.health.blueprint)
        self.register_blueprint(routes.index.blueprint)
        self.register_blueprint(routes.tags.blueprint)
        self.register_blueprint(routes.vote.blueprint)
        self.register_blueprint(routes.comment.blueprint)
        self.register_blueprint(routes.stats.blueprint)
        self.register_blueprint(routes.editor.blueprint)
        self.register_blueprint(routes.groups.blueprint)
        self.register_blueprint(routes.auth.blueprint)
        KnowledgeAuthProvider.register_auth_provider_blueprints(self)

        if self.config['DEBUG']:
            self.register_blueprint(routes.debug.blueprint)

        # Register error handler
        @self.errorhandler(500)
        def show_traceback(self):
            """ If LOCAL MODE: show the stack trace on a server error
                otherwise show a nice error template to the users
            """
            if current_app.config.get("DEBUG"):
                return render_template('traceback.html', info=traceback.format_exc()), 500
            else:
                return render_template('error.html')

        @self.errorhandler(404)
        def show_404(self):
            return render_template("404.html")

        @self.before_first_request
        def ensure_excluded_tags_exist():
            # For every tag in the excluded tags, create the tag object if it doesn't exist
            excluded_tags = current_app.config['EXCLUDED_TAGS']
            for tag in excluded_tags:
                Tag(name=tag)
            db_session.commit()

        @self.before_request
        def open_repository_session():
            if not request.path.startswith('/static'):
                current_repo.session_begin()

        @self.after_request
        def close_repository_session(response):
            if not request.path.startswith('/static'):
                current_repo.session_end()
            return response

        @self.context_processor
        def webediting_enabled():
            # TODO: Link this more to KnowledgeRepository capability and
            # configuration rather than a specific name.
            return {"webeditor_enabled": 'webposts' in current_repo.uris}

        @self.context_processor
        def inject_version():
            version = knowledge_repo.__version__
            version_revision = None
            if '_' in knowledge_repo.__version__:
                version, version_revision = version.split('_')
            return dict(version=version,
                        version_revision=version_revision,
                        last_index=time_since_index(human_readable=True),
                        last_index_check=time_since_index_check(human_readable=True))

        @self.template_global()
        def modify_query(**new_values):
            args = request.args.copy()

            for key, value in new_values.items():
                args[key] = value

            return u'{}?{}'.format(request.path, url_encode(args))

        @self.template_global()
        def pagination_pages(current_page, page_count, max_pages=5, extremes=True):
            page_min = int(max(1, current_page - math.floor(1.0 * max_pages // 2)))
            page_max = int(min(page_count, current_page + math.floor(1.0 * max_pages / 2)))

            to_acquire = max_pages - (page_max - page_min + 1)

            while to_acquire > 0 and page_min > 1:
                page_min -= 1
                to_acquire -= 1
            while to_acquire > 0 and page_max < page_count:
                page_max += 1
                to_acquire -= 1

            pages = list(range(page_min, page_max + 1))
            if extremes:
                if 1 not in pages:
                    pages[0] = 1
                if page_count not in pages:
                    pages[-1] = page_count
            return pages

        @self.template_filter('format_date')
        def format_date(date):
            """
            This will be a Jinja filter that string formats a datetime object.
            If we can't correctly format, we just return the object.
            :type date: Datetime
            :return: A string of the format of YYYY-MM-DD
            :rtype: String
            """
            try:
                return datetime.strftime(date, '%Y-%m-%d')
            except:
                return date

    @property
    def repository(self):
        return getattr(self, '_repository')

    @repository.setter
    def repository(self, repo):
        self._repository = repo

    @property
    def db(self):
        return sqlalchemy_db

    @property
    def _alembic_config(self):
        if not hasattr(self, 'extensions') or 'migrate' not in self.extensions:
            raise RuntimeError("KnowledgeApp has not yet been configured. Please instantiate it via `get_app_for_repo`.")
        migrations_path = os.path.join(os.path.dirname(__file__), "migrations")
        return self.extensions['migrate'].migrate.get_config(migrations_path)

    def db_init(self):
        with self.app_context():
            # Create all tables
            sqlalchemy_db.create_all()

            # Stamp table as being current
            command.stamp(self._alembic_config, "head")
        return self

    @property
    def db_revision(self):
        with self.app_context():
            conn = self.db.engine.connect()

            context = MigrationContext.configure(conn)
            return context.get_current_revision()

    def db_upgrade(self):
        with self.app_context():
            command.upgrade(self._alembic_config, "head")
        return self

    def db_downgrade(self, revision):
        with self.app_context():
            command.downgrade(self._alembic_config, revision)
        return self

    def db_migrate(self, message, autogenerate=True):
        with self.app_context():
            command.revision(self._alembic_config, message=message, autogenerate=autogenerate)
        return self

    def db_update_index(self, check_timeouts=True, force=False, reindex=False):
        with self.app_context():
            update_index(check_timeouts=check_timeouts, force=force, reindex=reindex)
        return self

    def start_indexing(self):
        if not getattr(self, '_indexing_started', False):
            set_up_indexing_timers(self)
            self._indexing_started = True
        return self

    def check_thread_support(self, check_index=True, check_repositories=True):
        # If index database is an sqlite database, it will lock on any write action, and so breaks on multiple threads
        # Repository uris will break as above (but less often since they are not often written too), but will also
        # end up being a separate repository per thread; breaking consistency of presented content.

        if check_index:
            index_db = self.config['SQLALCHEMY_DATABASE_URI']
            if index_db.startswith('sqlite://'):
                return False

        if check_repositories:
            for uri in self.repository.uris.values():
                if uri.startswith('sqlite://') or ':memory:' in uri:
                    return False

        return True



from abc import abstractmethod
from builtins import object

from flask_principal import identity_changed, Identity
from flask_login import login_user
from future.utils import with_metaclass
from flask import redirect, current_app, Blueprint, url_for, session

from .utils.auth import prepare_user
from ..utils.registry import SubclassRegisteringABCMeta


class KnowledgeAuthProvider(with_metaclass(SubclassRegisteringABCMeta, object)):
    _registry_keys = None

    @classmethod
    def register_auth_provider_blueprints(cls, app, prefix='/auth/login'):
        app.auth_providers = []
        for provider in app.config.get('AUTH_PROVIDERS', ['debug', 'oauth']):
            if not isinstance(provider, KnowledgeAuthProvider):
                provider = cls._get_subclass_for(provider.lower())(name=provider, app=app)
            app.register_blueprint(provider.blueprint, url_prefix='/'.join((prefix, provider.name)))
            app.auth_providers.append(provider)

    def __init__(self, name, app=None, **kwargs):
        self.name = name
        self.app = app
        self.blueprint = Blueprint('auth_provider.' + self.name, __name__, template_folder='../templates')
        self.prepare_blueprint(self.blueprint)
        self.init(**kwargs)

    def init(self):
        pass

    def prepare_blueprint(self, blueprint):
        blueprint.add_url_rule('/', view_func=self.prompt, methods=['GET'])
        blueprint.add_url_rule('/authorize', view_func=self.authorize, methods=['GET', 'POST'])
        return blueprint

    @property
    def link_text(self):
        return "Sign in using {}".format(self.name.capitalize())

    @property
    def icon_uri(self):
        return url_for('static', filename='images/auth/{}.png'.format(self.name))

    def authorize(self):
        user = self.get_user()
        if user is None:
            raise RuntimeError("No such user or invalid credentials")
        assert self.validate(user)
        self._perform_login(user)
        return redirect(session.get('requested_url') or url_for('index.render_feed'))

    def validate(self, user):
        return True

    def _perform_login(self, user):
        user = prepare_user(user)
        login_user(user)

        # Notify flask principal that the identity has changed
        identity_changed.send(current_app._get_current_object(),
                              identity=Identity(user.id))

    @abstractmethod
    def prompt(self):
        pass

    @abstractmethod
    def get_user(self):
        pass



import datetime
# ---------------------------------------------------
# Host configuration
# ---------------------------------------------------

# The server name is used by Flask to limit access to the
# served content to request to a particular domain. It
# is also used by some authentication providers (in particular
# OAuth providers) to advertise callback providers. If
# not provided, it is assumed in these contexts to be
# 'localhost:7000'. Be sure to specify this before deploying
# into production.
SERVER_NAME = None

# The knowledge repository uses the secret key to sign user
# sessions. If not specified, a unique secret key will be
# generated every time the server starts up. If hosting
# in a multi-server environment, or you want sessions
# to persist accross server restarts, set this to something
# static.
SECRET_KEY = None

# Set DEPLOY_HTTPS to True if you want to enable encrypted
# communication with Flask. When enabled, you must provide
# your ssl certificate, which consists of a .crt and .key
# file.
# Note: Even if you set DEPLOY_HTTPS to True, you still need
# to set the port to 443 manually.
DEPLOY_HTTPS = False
SSL_CERT = {
    'cert': '/path/to/cert',
    'key': '/path/to/key'
}

# ---------------------------------------------------
# Debug configuration
# ---------------------------------------------------
DEBUG = False

# ---------------------------------------------------
# Database configuration
# ---------------------------------------------------
SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
SQLALCHEMY_ECHO = False
SQLALCHEMY_TRACK_MODIFICATIONS = False

# Should the database tables be automatically created
DB_AUTO_CREATE = True

# Should the database be automatically migrated when updates exist
# Note: This is True by default if this configuration is not applied,
# under the assumption that if you are concerned about this file
# you are probably interested in minimising risk to stability and handling
# database upgrades manually. Manual database migrations can be
# performed using `knowledge_repo --repo <> db_upgrade ...`.
DB_AUTO_UPGRADE = False


# ---------------------------------------------------
# Authentication configuration
# ---------------------------------------------------
# Authentication providers allow users to sign into the Knowledge Repo
# in a variety of different ways. You can create your own subclass of
# `KnowledgeAuthProvider` and add either the instance or identifier
# used for that class below.
# By default, the knowledge repo offers:
# ['debug', 'oauth2', 'bitbucket', 'github', 'google']
AUTH_PROVIDERS = []

# If you are going to use a OAuth provider, you will need to specify client ids
# and private tokens. This can be done by instantiating instances of
# `OAuth2Provider` and adding them to the above list, or by specifying OAuth
# connection properties as demonstrated below for the GitHub authenticator.
# OAUTH_GITHUB_CLIENT_ID = '<client id>'
# OAUTH_GITHUB_CLIENT_SECRET = '<client secret>'

# To configure a generic OAuth provider that is not one of the presets
# provided, you may use the provider 'oauth2' which creates an empty,
# unconfigured OAuth2Provider. You must then override its configuration.
# For example, for a self-managed Gitlab CE instance at gitlab.example.com:

# OAUTH_OAUTH2_BASE_URL = 'https://gitlab.example.com/api/v4/'
# OAUTH_OAUTH2_AUTHORIZATION_URL = 'https://gitlab.example.com/oauth/authorize'
# OAUTH_OAUTH2_TOKEN_URL = 'https://gitlab.example.com/oauth/token'
# OAUTH_OAUTH2_AUTO_REFRESH_URL = 'https://gitlab.example.com/oauth/token'
# OAUTH_OAUTH2_SCOPES = 'api'
# OAUTH_OAUTH2_USER_INFO_ENDPOINT = 'user'
# OAUTH_OAUTH2_USER_INFO_MAPPING = {
#     'identifier': 'username',
#     'name': 'name',
#     'avatar_uri': 'avatar_url'
# }
# OAUTH_OAUTH2_VERIFY_SSL_CERTS = '/path/to/certs/my.ca-bundle'
# OAUTH_OAUTH2_CLIENT_ID = '<client id>'
# OAUTH_OAUTH2_CLIENT_SECRET = '<client secret>'

# The configuration OAUTH_<name>_VERIFY_SSL_CERTS is what is passed to the
# 'verify' parameter in the Requests module, and can be used to disable
# HTTPS verification (not recommended) or provide a custom CA bundle. See:
# http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification

# You may also override the .validate() method of a KnowledgeAuthProvider
# to perform an additional validation step before authenticating a user.
# The following example checks whether a user has access to the git remote
# of the local Knowledge Repository:

# def OAUTH_OAUTH2_VALIDATE(provider, user):
#
#     if provider.app.repository.git_has_remote:
#
#         url_parts = (
#             provider.app.repository.git_remote.url.split(':')
#             )
#
#         url_subparts = url_parts[1].split('/')
#
#         if url_parts[0] == "git@gitlab.example.com":
#             git_project = (
#                 url_subparts[0] + "%2F" + url_subparts[1].split('.')[0])
#         elif (
#             url_parts[0] == "https"
#             and url_subparts[2] == "gitlab.example.com"
#         ):
#             git_project = (
#                 url_subparts[3] + "%2F" + url_subparts[4].split('.')[0])
#         else:
#             provider.app.logger.warning(
#                 "User validation failed: unexpected git remote url ["
#                 + provider.app.repository.git_remote.url + "]")
#             return False
#
#         user_validate_url = provider.base_url + "projects/" + git_project
#
#         resp = provider.oauth_client.get(
#             user_validate_url,
#             verify=OAUTH_OAUTH2_VERIFY_HTTPS)
#
#         if resp.status_code == 200:
#             return True
#         else:
#             provider.app.logger.warning(
#                 "User validation failed: validation URL ["
#                 + user_validate_url + "] returned HTTP status ["
#                 + str(resp.status_code) + "]")

# You can also forgo a fully-fledged sign in process for users by hosting the
# knowledge repository behind a proxy server that pre-authenticates users, and
# adds the appropriate user identifier to the http headers of the request. If
# enabled below, then they take precedence over any other forms of
# authentication. If the call to `AUTH_MAP_REQUEST_HEADERS` results in a null
# user identifier, then the authentication flow will fall back to use any of the
# providers specified above.
AUTH_USE_REQUEST_HEADERS = False


# If using headers to authenticate, the following function should be implemented
# to transform a dictionary of headers into a dictionary of user attributes.
# Currently only 'identifier', 'avatar_uri', 'name' and 'email' are supported.
# If this method returns `None`, or `identifier` is not supplied, then the
# authorization flow will fall back to other authentication methods.
def AUTH_MAP_REQUEST_HEADERS(headers):
    return {
        # 'identifier': None,
        # 'avatar_uri': None,
        # 'name': None,
        # 'email': None
    }


# The following AUTH_USER_IDENTIFIER* configuration keys are deprecated and
# will be removed in v0.9.
AUTH_USER_IDENTIFIER_REQUEST_HEADER = None


def AUTH_USER_IDENTIFIER_REQUEST_HEADER_MAPPING(identifier):
    return identifier


# If the server desires to modify the attributes of the `User` object associated with
# users logged in via any of the above authentication providers, it can do so via
# this configuration key. This function will be run once at user login (if using
# an `AuthenticationProvider`, and then at most once during any caching lifetime
# period (as specified below). Note that attributes collected via
# `AuthenticationProvider`s will not be updated after initial login (user must
# relogin in order to reset those attributes).
def AUTH_USER_ATTRIBUTE_SETTER(user):
    return user


# The time to wait before re-checking user attributes with the above function
# for users logged in via request headers.
AUTH_USER_ATTRIBUTE_CACHE_LIFETIME = 24 * 60 * 60  # 1 day

# Once a user is logged in via an authentication provider, they will remain
# logged in via the use of cookies. By default, this cookie will last one year.
# This is managed by `flask_login`, but is copied here for convenience.
# For other options regarding sessions, please refer to:
# https://flask-login.readthedocs.io/en/latest/#cookie-settings
REMEMBER_COOKIE_DURATION = datetime.timedelta(days=365)

# ---------------------------------------------------
# Policy configuration
# ---------------------------------------------------
# This section configures various policy related to access control.

# Should anonymous users be able to view the post indices
POLICY_ANONYMOUS_VIEW_INDEX = True

# Should anonymous users be able to view the content of posts
POLICY_ANONYMOUS_VIEW_POST = True

# Should anonymous users be able to view overall statistics
POLICY_ANONYMOUS_VIEW_STATS = True

# Should anonymous users be able to download posts (or their source)
POLICY_ANONYMOUS_DOWNLOADS = False


# ---------------------------------------------------
# Repository configuration
# ---------------------------------------------------
# You may specify a function `prepare_repo` which configures
# the repository upon which this server is running. This
# takes place after the repository has been instantiated
# and before the server is able to serve requests. It is
# possible to do anything to the repository, including
# substituting the repository for another one.
# By default, repositories manage their own configurations,
# but this can be risky as they may run arbitrary python code,
# which opens a vector for malicious users to compromise
# the server. If you want to avoid this risk, pass
# the '--safe' (TODO!) option to `knowledge_repo` config and
# manually configure the repository here.
# For example, if your server instance is sitting atop
# a meta-repository, it may make sense to update the meta-repository
# configuration with that of one of its children.
def prepare_repo(repo):
    return repo


# ---------------------------------------------------
# Repository Indexing configuration
# ---------------------------------------------------
# The Knowedge Repo updates the index of available posts on a regular basis.
# If the database is not thread-safe (i.e. in the case of SQLite), then the
# index will be updated on the main thread before every request that is more
# than `INDEX_INTERVAL` seconds after the last sync completed. Otherwise,
# indexing will occur every `INDEX_INTERVAL` seconds after the previous sync.
# Syncing is designed to be compatible with multiple instances of the Knowledge
# Repo connected to the same database, accross multiple machines and/or
# processes; and so a global indexing lock is employed. When a sync begins,
# a sync lock is put in place and the responsible process is considered to be
# the primary agent responsible for syncing until its last update is longer than
# `INDEXING_TIMEOUT` seconds, whereby the lock is ceded to the next requesting
# process. Note that `INDEXING_TIMEOUT` must be larger than `INDEXING_INTERVAL`
# or strange things might begin to happen.
INDEXING_INTERVAL = 5 * 60  # 5 minutes
INDEXING_TIMEOUT = 10 * 60  # 10 minutes

# Whether an index operation should update repositories
INDEXING_UPDATES_REPOSITORIES = True

# Whether repositories should be updated even without a sync lock (in which case
# the repositories will be updated on the sync timers, even if the relevant
# process/thread does not have a lock on updating the index). This is useful in
# context of multiple Knowledge Repo servers working together to serve the
# repositories across multiple machines, which each require repository syncing.
# Disable this if (for some reason) you have multiple Knowledge Repo servers
# running on the same machine, and you want to avoid potential clashes. This
# key is ignored if `INDEXING_UPDATES_REPOSITORIES` is False
INDEXING_UPDATES_REPOSITORIES_WITHOUT_LOCK = True

# In some cases you may want to disable indexing entirely, which is currently
# only ever used by the Knowledge Post previewer. Disabling the index means that
# posts will not be discoverable, but if know the path in the repository you can
# view the post with a direct link.
INDEXING_ENABLED = True

# ---------------------------------------------------
# Flask Mail Configuration
# Refer to https://pythonhosted.org/flask-mail/
# Unless specified, upstream defaults are used as indicated
# provided that MAIL_SERVER is defined.
# ---------------------------------------------------
# MAIL_SERVER = 'localhost'  # default = 'localhost'
# MAIL_PORT = 25  # default = 25
# MAIL_USE_TLS = False  # default = False
# MAIL_USE_SSL = False  # default = False
# MAIL_DEBUG = False  # default = app.debug
# MAIL_USERNAME = None  # default = None
# MAIL_PASSWORD = None  # default = None
# MAIL_DEFAULT_SENDER = None  # default = None
# MAIL_MAX_EMAILS = None  # default = None
# MAIL_SUPPRESS_SEND = False  # default = app.testing
# MAIL_ASCII_ATTACHMENTS = False  # default = False


# --------------------------------------------------
# Web Editor Configuration
# --------------------------------------------------
# The web editor can be limited to editing posts under
# a limited set of parent directories by setting
# WEB_EDITOR_PREFIXES to a list of supported path prefixes.
# e.g. ['webposts', 'projects']
WEB_EDITOR_PREFIXES = ['webposts']


# ---------------------------------------------------
# Tag configuration
# ---------------------------------------------------
# Posts with certain tags can be excluded from showing up
# in the app. This can be useful for security purposes
EXCLUDED_TAGS = ['private']



import sys
import importlib
import pkgutil


def import_submodules(package_name):
    """ Import all submodules of a module, recursively

    :param package_name: Package name
    :type package_name: str
    :rtype: list[types.ModuleType]
    """
    package = sys.modules[package_name]
    submodules = []
    for _, name, _ in pkgutil.walk_packages(package.__path__):
        importlib.import_module(package_name + '.' + name)
        submodules.append(name)
    return submodules



import logging

from flask import request, render_template, Blueprint
from sqlalchemy import and_

from ..proxies import db_session
from ..models import Group, Post, PageView, User, assoc_group_user

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


blueprint = Blueprint('groups', __name__,
                      template_folder='../templates', static_folder='../static')


@blueprint.route('/post_groups', methods=['GET'])
@PageView.logged
def groups_for_posts():
    """Given a post_id, return a page where you see a list of groups,
       and can view the group members and add people to groups
    """
    post_id = request.args.get('post_id', 0)
    post = (db_session.query(Post)
                      .filter(Post.id == post_id)
                      .first())
    return render_template('post_groups.html',
                           groups=post.groups)


@blueprint.route('/view_group_users', methods=['GET'])
@PageView.logged
def view_group_users():
    """Given a group_id, return a page with a list of group members"""
    group_id = request.args.get('group_id', 0)
    group = (db_session.query(Group)
                       .filter(Group.id == group_id)
                       .first())
    return render_template('view_group_users.html',
                           group=group)


@blueprint.route('/add_group_users', methods=['GET'])
@PageView.logged
def add_group_users():
    """Given a group_id, show a page where users can be added"""
    group_id = request.args.get('group_id', 0)
    group = (db_session.query(Group)
                       .filter(Group.id == group_id)
                       .first())
    return render_template('add_group_users.html',
                           group=group)


@blueprint.route('/add_users_to_group', methods=['POST', 'GET'])
@PageView.logged
def add_users_to_group():
    group_id = request.args.get('group_id', 0)
    users = request.get_json()
    user_objs = []
    for user in users:
        username = user.strip()
        user_obj = (db_session.query(User)
                              .filter(User.identifier == username)
                              .first())
        if not user_obj:
            user_obj = User(identifier=username)
            db_session.add(user_obj)
        user_objs.append(user_obj)

    group_obj = (db_session.query(Group)
                           .filter(Group.id == group_id)
                           .first())
    group_obj.users = user_objs
    db_session.commit()
    return ""


@blueprint.route('/delete_user_from_group', methods=['POST', 'GET'])
@PageView.logged
def delete_user_from_group():
    group_id = request.args.get('group_id', 0)
    user_id = request.get_json()
    user_obj = (db_session.query(User)
                          .filter(User.id == user_id)
                          .first())
    delete_query = (assoc_group_user.delete()
                                    .where(and_(assoc_group_user.c.group_id == group_id,
                                                assoc_group_user.c.user_id == user_obj.id)))
    db_session.execute(delete_query)
    db_session.commit()
    return ""


@blueprint.route('/all_groups', methods=['POST', 'GET'])
@PageView.logged
def all_groups():
    groups = db_session.query(Group).all()
    return render_template('all_groups.html', groups=groups)


@blueprint.route('/add_groups', methods=['POST', 'GET'])
@PageView.logged
def add_groups():
    groups = request.get_json()
    for group_name in groups:
        group = Group(name=group_name)
        db_session.add(group)
    db_session.commit()
    return ""
